#ifndef OCCUPANCY_GRID_UTILS_RAY_TRACER_H
#define OCCUPANCY_GRID_UTILS_RAY_TRACER_H

#include <occupancy_grid_utils/impl/ray_trace_iterator.h>
#include <nav_msgs/OccupancyGrid.h>
#include <geometry_msgs/Pose.h>

namespace occupancy_grid_utils
{

/************************************************************
 * Basic ray tracing
 ************************************************************/

typedef std::pair<RayTraceIterator, RayTraceIterator> RayTraceIterRange;

/// \brief Returns an iterator range over the cells on the 
/// line segment between two points (inclusive).  
/// \param project_onto_grid If true, \a p2 may be off the grid, in which case the ray
/// stops right before falling off the grid
/// \throws PointOutOfBoundsException if \a p1 is off grid
/// \throws PointOutOfBoundsException if \a p2 is off grid and project_onto_grid is false
/// \retval range models
/// <a href="http://www.boost.org/doc/libs/1_42_0/libs/range/doc/range.html#forward_range">Forward range</a> 
/// with reference type const Cell&, i.e., it's a pair of iterators that you can use 
/// in for loops, std::for_each, etc.
RayTraceIterRange rayTrace (const nav_msgs::MapMetaData& info, const geometry_msgs::Point& p1, 
                            const geometry_msgs::Point& p2, bool project_onto_grid=false);



/************************************************************
 * Overlaying point clouds
 ************************************************************/

/// \brief Represents a cloud of points (in a sensor frame), together with the pose (in the reference frame)
/// of the sensor that detected them
struct LocalizedCloud
{
  LocalizedCloud ();
  LocalizedCloud (const geometry_msgs::Pose& sensor_pose, const std::vector<geometry_msgs::Point>& points);
  geometry_msgs::Pose sensor_pose;
  std::vector<geometry_msgs::Point> points;
};


/// Default occupancy threshold for overlayClouds
const double DEFAULT_OCCUPANCY_THRESHOLD=0.1;

/// Default max_distance for overlayClouds
const double DEFAULT_MAX_DISTANCE=10.0;

/// Default min_pass_through for overlayClouds 
const double DEFAULT_MIN_PASS_THROUGH=2;

/// \brief Overlay a set of point clouds into a grid
///
/// Uses \a info to generate an empty occupancy grid.  Then generates the rays between each sensor and 
/// corresponding points. Sets a given cell to UNKNOWN if fewer than \a min_pass_through rays pass 
/// through it, else UNOCCUPIED if fewer than an \a occupancy_threshold fraction of the rays passing 
/// through it stop at it, else OCCUPIED.
/// 
/// If a point on a cloud is off the grid and within \a max_distance, we clear all points on the ray 
/// to it that are still on the grid.
/// 
/// Points that are more than max_distance from the corresponding sensor are skipped over whether or
/// not they're on the grid
nav_msgs::OccupancyGrid overlayClouds (const nav_msgs::MapMetaData& info, 
                                       const std::vector<LocalizedCloud>& clouds,
                                       double occupancy_threshold=DEFAULT_OCCUPANCY_THRESHOLD, 
                                       double max_distance=DEFAULT_MAX_DISTANCE,
                                       double min_pass_through=DEFAULT_MIN_PASS_THROUGH);




} // namespace occupancy_grid_utils

#endif // include guard
