/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * \file
 * 
 * Implements overlayClouds
 *
 * \author Bhaskara Marthi
 */

#include <occupancy_grid_utils/ray_tracer.h>
#include <ros/assert.h>
#include <tf/transform_datatypes.h>
#include <boost/foreach.hpp>
#include <boost/bind.hpp>
#include <boost/ref.hpp>
#include <boost/optional.hpp>

namespace occupancy_grid_utils
{

namespace gm=geometry_msgs;
namespace nm=nav_msgs;

using boost::bind;
using boost::ref;
using std::vector;

// Constructors
LocalizedCloud::LocalizedCloud (const gm::Pose& sensor_pose, const vector<gm::Point>& points) :
  sensor_pose(sensor_pose), points(points)
{}

LocalizedCloud::LocalizedCloud () 
{}
  


gm::Point transformPt (const btTransform& trans, const gm::Point& p)
{
  tf::Point pt;
  tf::pointMsgToTF(p, pt);
  gm::Point transformed;
  tf::pointTFToMsg(trans*pt, transformed);
  return transformed;
}

double euclideanDistance (const gm::Point& p1, const gm::Point& p2)
{
  const double dx=p1.x-p2.x;
  const double dy=p1.y-p2.y;
  const double dz=p1.z-p2.z;
  return sqrt(dx*dx + dy*dy + dz*dz);
}


// This is our policy for computing the occupancy of a cell based on hit and pass through counts
int8_t determineOccupancy (const unsigned hit_count, const unsigned pass_through_count, 
                           const double occupancy_threshold, const double min_pass_through)
{
  if (pass_through_count < min_pass_through)
    return UNKNOWN;
  else if (hit_count > pass_through_count*occupancy_threshold)
    return OCCUPIED;
  else 
    return UNOCCUPIED;
}


// Main function
nm::OccupancyGrid overlayClouds (const nm::MapMetaData& info, const vector<LocalizedCloud>& clouds,
                                 const double occupancy_threshold, const double max_distance,
                                 const double min_pass_through)
{
  const index_t num_cells = info.width*info.height;
  ROS_ASSERT(min_pass_through>0);
  ROS_ASSERT(num_cells>0);

  vector<unsigned> hit_counts(num_cells);
  vector<unsigned> pass_through_counts(num_cells);

  /// Iterate over clouds
  BOOST_FOREACH (const LocalizedCloud& loc_cloud, clouds) {
    const gm::Point& sensor_pos = loc_cloud.sensor_pose.position;

    // Skip when this sensor is not on the grid
    if (!withinBounds(info, sensor_pos)) {
      ROS_INFO_NAMED ("overlay", "Sensor pose %.2f, %.2f was not on grid; skipping", 
                      sensor_pos.x, sensor_pos.y);
      continue;
    }

    // Transform points to world frame
    btTransform sensor_to_world;
    tf::poseMsgToTF(loc_cloud.sensor_pose, sensor_to_world);
    vector<gm::Point> transformed_points(loc_cloud.points.size());
    transform(loc_cloud.points.begin(), loc_cloud.points.end(), transformed_points.begin(),
              bind(transformPt, ref(sensor_to_world), _1));

    // Iterate over points in this cloud
    BOOST_FOREACH (const gm::Point& p, transformed_points) {
      
      // Skip point unless it's within the distance threshold
      if (euclideanDistance(sensor_pos, p) < max_distance) {
        ROS_DEBUG_NAMED ("overlay", "Ray tracing to point %.2f, %.2f", p.x, p.y);
        boost::optional<index_t> last_ind;
        BOOST_FOREACH (const Cell& c, rayTrace(info, sensor_pos, p, true)) {
          last_ind = cellIndex(info, c);
          pass_through_counts[*last_ind]++;
          ROS_DEBUG_NAMED ("overlay", "Pass-through counts for %d, %d are now %u", c.x, c.y, 
                           pass_through_counts[*last_ind]);
        }
        ROS_ASSERT(last_ind);

        // If the last cell equals the point (i.e., point is not off the grid), update hit counts
        const Cell last_cell = indexCell(info, *last_ind);
        if (last_cell == pointCell(info, p)) {
          hit_counts[*last_ind]++;
          ROS_DEBUG_NAMED ("overlay", "Hit counts for %d, %d are now %u", last_cell.x, last_cell.y, 
                           hit_counts[*last_ind]);
        }
        else
          ROS_ASSERT (!withinBounds(info, pointCell(info, p)));
      }
    }
  }
  
  /// Construct an occupancy grid based on the counts
  nm::OccupancyGrid grid;
  grid.info = info;
  grid.data.resize(num_cells);
  transform(hit_counts.begin(), hit_counts.end(), pass_through_counts.begin(), 
            grid.data.begin(), bind(determineOccupancy, _1, _2, occupancy_threshold, min_pass_through));
  return grid;
}




} // namespace occupancy_grid_utils
