/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */


/**
 * \file
 *
 * Define a base class for constraint generators
 * 
 * \author Bhaskara Marthi
 */

#ifndef POSE_GRAPH_CONSTRAINT_GENERATOR_H
#define POSE_GRAPH_CONSTRAINT_GENERATOR_H

#include <pose_graph/pose_graph.h>
#include <geometry_msgs/PoseStamped.h>
#include <boost/shared_ptr.hpp>
#include <vector>
#include <map>

// Avoiding include all of ros into constraint generators that include this file
namespace ros
{
class NodeHandle;
}

namespace pose_graph
{


using geometry_msgs::Pose;


/************************************************************
 * Abstract constraint generator
 ************************************************************/


struct NodeConstraint
{
  NodeConstraint (const NodeId target, const PoseConstraint& constraint) :
    target(target), constraint(constraint) {}
  NodeId target;
  PoseConstraint constraint;
  EIGEN_MAKE_ALIGNED_OPERATOR_NEW
};

typedef std::vector<NodeConstraint, Eigen::aligned_allocator<NodeConstraint> > NodeConstraintVector;

/// Subclasses of this must define a method for getConstraints
/// They may assume getConstraints is called once each time a new node is added
class ConstraintGenerator
{
public:
  virtual ~ConstraintGenerator () {};

  /// \brief Return a set of constraints given the current state of the graph
  /// the identity of the new node, and current optimized poses
  virtual NodeConstraintVector getConstraints (const PoseGraph& graph, NodeId new_node, 
                                               const NodePoseMap& optimized_poses) = 0;
};

typedef boost::shared_ptr<ConstraintGenerator> ConstraintGenPtr;


// Create a constraint generator based on parameters in a ros namespace
ConstraintGenPtr createConstraintGenerator (const ros::NodeHandle& nh);

} // namespace pose_graph

#endif // include guard
