/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */


/**
 * \file
 *
 * Defines KartoLaserConstraints
 *
 * \author Bhaskara Marthi
 */

#ifndef POSE_GRAPH_KARTO_LASER_CONSTRAINTS_H
#define POSE_GRAPH_KARTO_LASER_CONSTRAINTS_H

#include <pose_graph/constraint_generator.h>
#include <pose_graph/sequential_scan_matcher.h>
#include <pose_graph/loop_scan_matcher.h>
#include <ros/ros.h>
#include <boost/thread.hpp>

namespace pose_graph
{

typedef boost::shared_ptr<SequentialScanMatcher> SeqMatcherPtr;
typedef boost::shared_ptr<LoopScanMatcher> LoopMatcherPtr;

/************************************************************
 * implements ConstraintGenerator; generates both
 * "near-chain", and loop-closure constraints, all based
 * on the Karto scan matcher
 ************************************************************/

class KartoLaserConstraints : public ConstraintGenerator
{
public:

  KartoLaserConstraints (double max_seq_link_length, unsigned min_seq_chain_length, 
                         double min_seq_link_response, double max_loop_link_length,
                         unsigned min_loop_chain_length, double min_loop_closure_response);

  NodeConstraintVector getConstraints (const PoseGraph& graph, const NodeId new_node,
                                       const NodePoseMap& optimized_poses);

private:

  void scanCallback(const sensor_msgs::LaserScan::ConstPtr& scan);
  NodeConstraintVector scanOdomConstraints (const PoseGraph& graph, const NodeId new_node,
                                       const NodePoseMap& optimized_poses);
  NodeConstraintVector loopConstraints (const PoseGraph& graph, const NodeId new_node,
                                        const NodePoseMap& optimized_poses);

  std::vector<Chain> nearLinkedChains (const PoseGraph& graph, const Pose& pose, LaserPtr scan);

  const double max_sequential_link_length_;
  const unsigned min_sequential_chain_length_;
  const double min_sequential_link_response_;
  const double max_loop_link_length_;
  const unsigned min_loop_chain_length_;
  const double min_loop_closure_response_;

  boost::mutex mutex_;
  ros::NodeHandle nh_;
  ros::Subscriber scan_sub_;
  tf::TransformListener tf_;
  SeqMatcherPtr seq_matcher_;
  LoopMatcherPtr loop_matcher_;

  sensor_msgs::LaserScan::ConstPtr last_scan_;
};



} // namespace pose_graph

#endif // include guard
