/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */


/************************************************************ 
 * Namespace and includes
 ************************************************************/

#ifndef POSE_GRAPH_EXCEPTION_H
#define POSE_GRAPH_EXCEPTION_H

#include <pose_graph/pose_graph.h>
#include <boost/format.hpp>
#include <stdexcept>
#include <string>

namespace pose_graph
{

using boost::format;
using std::string;


/// A base class for all pose_graph exceptions; provides a handy boost::format parent constructor
class PoseGraphException: public std::logic_error
{
public:
  PoseGraphException (const format& error_string) : std::logic_error(error_string.str()) {};
  PoseGraphException (const char* str) : std::logic_error(str) {};
};



/// Exception for unknown NodeId
struct UnknownNodeIdException: public PoseGraphException
{
  UnknownNodeIdException (const NodeId id): 
    PoseGraphException(format("Unknown node id %1%") % id), id(id) {}
  const NodeId id;
};

/// Exception for unknown EdgeId
struct UnknownEdgeIdException: public PoseGraphException
{
  UnknownEdgeIdException (const EdgeId id): 
    PoseGraphException(format("Unknown edge id %1%") % id), id(id) {}
  const EdgeId id;
};


/// Exception for trying to add a NodeId that already exists
struct DuplicateNodeIdException: public PoseGraphException
{
  DuplicateNodeIdException (const NodeId id):
    PoseGraphException(format("Node %1% already exists") % id), id(id) {}
  const NodeId id;
};


/// Exception for trying to add an EdgeId that already exists
struct DuplicateEdgeIdException: public PoseGraphException
{
  DuplicateEdgeIdException (const EdgeId id):
    PoseGraphException(format("Edge %1% already exists") % id), id(id) {}
  const EdgeId id;
};


/// Exception for failing to open a file
struct FileOpenException: public PoseGraphException
{
  FileOpenException (const string& filename):
    PoseGraphException(format("Failed to open file %1%") % filename) {}
};

/// Exception for incorrect format of saved graph
struct FileFormatException: public PoseGraphException
{
  FileFormatException (const string& filename): 
    PoseGraphException(format("%1% is not a validly formatted pose graph file") % filename) {}
};

/// Exception for a topic that hasn't been registered with a pose graph
struct UnknownTopicException: public PoseGraphException
{
  UnknownTopicException (const string& topic): 
    PoseGraphException(format("%1% is not a known topic in this pose graph") % topic) {}
};

/// Exception for not finding a ros parameter
struct NonexistentParamException: public PoseGraphException
{
  NonexistentParamException (const string& name) :
    PoseGraphException(format ("Ros parameter %1% was not set") % name) {}
};


} // namespace

#endif // POSE_GRAPH_EXCEPTION_H
