/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */


/**
 * \file
 *
 * Contains definitions of geometric objects stored in the pose graph
 */


#ifndef POSE_GRAPH_GEOMETRY_H
#define POSE_GRAPH_GEOMETRY_H

#include <Eigen/Dense>
#include <iostream>
#include <geometry_msgs/Pose.h>
#include <geometry_msgs/Point32.h>
#include <geometry_msgs/Pose2D.h>
#include <sensor_msgs/LaserScan.h>

namespace pose_graph
{

using std::ostream;

typedef Eigen::Matrix<double, 6, 6> PrecisionMatrix;

/// A single pose constraint (including precision matrix) between two nodes
struct PoseConstraint
{
  PoseConstraint ();
  PoseConstraint (const Eigen::Vector3d& translation, const Eigen::Quaterniond& rotation);
  PoseConstraint (const Eigen::Vector3d& translation, const Eigen::Quaterniond& rotation, const PrecisionMatrix& precision);

  Eigen::Vector3d translation;
  Eigen::Quaterniond rotation;
  PrecisionMatrix precision;

  EIGEN_MAKE_ALIGNED_OPERATOR_NEW
};

ostream& operator<< (ostream& str, const PoseConstraint& constraint);

/// Norm of a vector
double length (const Eigen::Vector3d& v);

/// Norm of mean translation part of a Pose constraint
double length (const PoseConstraint& constraint);

/// Given a pose constraint, return the expected relative pose
geometry_msgs::Pose getPose (const PoseConstraint& c);

// Given a transform between two frames, get a PoseConstraint.  
// We set the precision matrix to the identity.
PoseConstraint makeConstraint (const Eigen::Transform3d& trans);

// Get distance between two points
double euclideanDistance (const geometry_msgs::Point& p1, const geometry_msgs::Point& p2);

// Get distance between two poses, ignoring angles
double euclideanDistance (const geometry_msgs::Pose2D& p1, const geometry_msgs::Pose2D& p2);

// Transform between point32 and point
geometry_msgs::Point convertToPoint (const geometry_msgs::Point32& p);

// Barycenter of a scan
geometry_msgs::Point computeBarycenter (const sensor_msgs::LaserScan& scan);



} // namespace pose_graph

#endif // POSE_GRAPH_GEOMETRY_H
