/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */


/**
 * \file
 *
 * Contains the interface for the (ros-independent) pose graph data structures
 *
 * \author Bhaskara Marthi
 */



#ifndef POSE_GRAPH_POSE_GRAPH_H
#define POSE_GRAPH_POSE_GRAPH_H

#include "geometry.h"
#include <pose_graph/LocalizedCloud.h>
#include <boost/scoped_ptr.hpp>
#include <boost/optional.hpp>
#include <set>
#include <vector>

namespace pose_graph {

/************************************************************ 
 * Externally used typedefs (see also geometry.h)
 ************************************************************/

// Ids used by the outside world to refer to nodes and edges
// Can be assumed to be unsigned integer types
typedef uint32_t NodeId;
typedef uint32_t EdgeId;

/// Result of a shortest path query
struct ShortestPathResult
{
  /// Whether a path was found; other fields only valid if this is true
  bool path_found;
  
  /// Nodes in path starting at source, ending at destination
  std::vector<NodeId> nodes;

  /// Edges in path from source to destination; not currently filled in
  std::vector<EdgeId> edges;

  /// Cost of path
  double cost;
};

/// Maps Node ids to poses
typedef std::map<NodeId, geometry_msgs::Pose> NodePoseMap;


/************************************************************
 * Main pose graph data structure
 ************************************************************/

class PoseGraphImpl;

class PoseGraph
{
public:


  /****************************************
   * Creation
   ****************************************/
  
  /// Constructs an empty graph
  PoseGraph ();

  /// Copy constructor
  PoseGraph (const PoseGraph& g);

  /// Assignment operator
  PoseGraph& operator= (const PoseGraph& g);

  /// Destructor
  ~PoseGraph();
  
  
  /****************************************
   * Basic graph modification ops
   ****************************************/

  /// \brief Adds a node
  /// \return NodeId of new node
  /// \post The new node's initial pose estimate is the identity pose.
  NodeId addNode ();

  /// \brief Adds a node with a given id
  /// \throw DuplicateNodeIdException if already exists
  /// \post The new node's initial pose estimate is the identity pose.
  void addNode (NodeId id);

  /// \brief Adds an edge
  /// \param from tail of new edge
  /// \param to head of new edge
  /// \param constraint PoseConstraint expressing head pose in frame of tail pose
  /// \return EdgeId of new edge
  /// \throw UnknownNodeIdException if either id is not known
  EdgeId addEdge (NodeId from, NodeId to, const PoseConstraint& constraint);

  /// \brief Like addEdge, but with a given id
  /// \param from tail of new edge
  /// \param to head of new edge
  /// \param constraint PoseConstraint expressing head pose in frame of tail pose
  /// \param id Desired id for new edge
  /// \throw UnknownNodeIdException if either id is not known
  /// \throw DuplicateEdgeIdException if already exists
  void addEdge (NodeId from, NodeId to, const PoseConstraint& constraint, EdgeId id);

  /// \brief Sets the initial pose estimate attached to this node.
  /// \throw UnknownNodeIdException
  void setInitialPoseEstimate (NodeId n, const geometry_msgs::Pose& pose);

  /// \brief Gets the initial pose estimate attached to this node.
  /// \throw UnknownNodeIdException
  /// \retval A reference valid as long as the node exists
  const geometry_msgs::Pose& getInitialPoseEstimate (NodeId n) const;


  /****************************************
   * Basic const ops
   ****************************************/

  /// Gets set of node ids in graph
  std::set<NodeId> allNodes() const;

  /// Gets set of edge ids in graph
  std::set<EdgeId> allEdges() const;

  /// \brief Gets the constraint attached to an edge.  
  /// \retval A reference that is valid so long as edge is in graph.
  /// \throw UnknownEdgeIdException
  const PoseConstraint& getConstraint (EdgeId id) const;

  /// \brief Gets length of an edge
  /// \throw UnknownEdgeIdException
  double edgeLength (EdgeId id) const;

  /// \brief gets adjacent edges to a node
  /// \throws UnknownNodeIdException
  std::set<EdgeId> incidentEdges (NodeId n) const;

  /// \brief gets incident nodes to an edge, in the order (tail, head).
  /// \throws UnknownEdgeIdException
  std::pair<NodeId, NodeId> incidentNodes (EdgeId e) const;

  /// \brief Gets neighboring nodes of a node.
  /// \throws UnknownNodeIdException
  std::set<NodeId> neighbors (NodeId n) const;

  /// \brief Checks if a node id already exists
  bool nodeIdExists (NodeId id) const;

  /// \brief Checks if an edge id already exists
  bool edgeIdExists (EdgeId id) const;

  /// \brief Get subgraph over a specific set of nodes
  ///
  /// The returned graph consists of the given nodes, with the same ids
  /// and all edges between them.  It does not share any structure with 
  /// the original graph.
  PoseGraph subgraph (const std::set<NodeId>& nodes) const;


  /****************************************
   * Graph search
   ****************************************/

  /// \brief Computes shortest path using Dijkstra's algorithm
  /// \param src Source node id
  /// \param dest Destination node id
  /// \throws UnknownNodeIdException
  ShortestPathResult shortestPath(NodeId src, NodeId dest) const;

  /// \brief Nodes within a given radius of current one
  /// \param src Source node
  /// \param radius
  /// \throws UnknownNodeIdException
  /// \retval Set of nodes, including src, whose shortest path distance from src is <= radius
  std::set<NodeId> nearbyNodes (NodeId src, double radius) const;


  /****************************************
   * Point clouds attached to nodes
   ****************************************/

  /// Attach a point cloud to a node.  
  /// \throws UnknownNodeIdException
  /// \param The reference frame of the cloud is the node frame
  void attachCloud (NodeId id, const LocalizedCloud& cloud);

  /// \brief Get the cloud attached to a node.
  /// \retval A reference that's valid as long as the node's in the graph
  /// \throws UnknownNodeIdException
  /// \throws PoseGraphException (if no cloud found for this node)
  const LocalizedCloud& getCloud (NodeId id) const;

  /// Return true iff the node has an attached cloud
  /// \throws UnknownNodeIdException
  /// \warning This is temporary pending rosrecord changes.
  bool hasCloud (NodeId id) const;

  /// Return the string naming the frame attached to a given node id
  static std::string nodeFrameName (const NodeId id);
  
private:
  
  boost::scoped_ptr<PoseGraphImpl> impl_;

};


} //namespace pose_graph


#endif // POSE_GRAPH_POSE_GRAPH_H
