/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */


/**
 * \file
 *
 * Implementation class for PoseGraph
 */



#ifndef POSE_GRAPH_POSE_GRAPH_IMPL_H
#define POSE_GRAPH_POSE_GRAPH_IMPL_H

#include "pose_graph.h"

#define BOOST_NO_HASH
#include <boost/graph/adjacency_list.hpp>

#include <boost/graph/graph_traits.hpp>

namespace pose_graph {

using boost::listS;
using boost::multisetS;
using boost::undirectedS;
using boost::adjacency_list;
using boost::graph_traits;
using std::map;
using std::vector;
using std::set;
using std::pair;

/************************************************************
 * Graph-related typedefs
 ************************************************************/

// Data structure containing info about a node in the pose graph
struct NodeInfo
{
  NodeInfo(NodeId id) : id(id) { initial_pose_estimate.orientation.w = 1.0; }
  NodeInfo() : id(-1) {} // Only used temporarily by boost graph operations
  NodeId id;
  geometry_msgs::Pose initial_pose_estimate;
};

// Data structure containing info about an edge in the pose graph
struct EdgeInfo
{
  EdgeInfo(EdgeId id, NodeId from, NodeId to, double length);
  EdgeInfo(); // Only used temporarily by boost graph operations

  // Unique id for this edge
  const EdgeId id;

  // Edges are undirected in the underlying graph, but we distinguish
  // between the from and to node ids here.  The constraint attached
  // to an edge specifies the pose of the to node in terms of the from node.
  const NodeId from, to;

  // Expectation of length of edge (given Gaussian pose constraint)
  const double length;
};


// Map from edge id to pose constraint
typedef Eigen::aligned_allocator<pair<EdgeId, PoseConstraint> > Allocator;
typedef map<EdgeId, PoseConstraint, std::less<EdgeId>, Allocator> PoseConstraintMap;


// boost graph and related types
typedef adjacency_list<multisetS, listS, undirectedS, NodeInfo, EdgeInfo> PoseGraphAdjacencyList;
typedef graph_traits<PoseGraphAdjacencyList>::edge_descriptor PoseGraphEdge;
typedef graph_traits<PoseGraphAdjacencyList>::vertex_descriptor PoseGraphVertex;


/************************************************************
 * Main implementation data structure
 ************************************************************/


class PoseGraphImpl
{
public:


  /****************************************
   * Creation
   ****************************************/
  
  /// Constructs an empty graph
  PoseGraphImpl ();

  /// Copy constructor
  PoseGraphImpl (const PoseGraphImpl& g);

  /// Assignment operator
  PoseGraphImpl& operator= (const PoseGraphImpl& g);
  
  
  /****************************************
   * Basic graph modification ops
   ****************************************/

  /// \brief Adds a node
  /// \return NodeId of new node
  /// \post The new node's initial pose estimate is the identity pose.
  NodeId addNode ();

  /// \brief Adds a node with a given id
  /// \throw DuplicateNodeIdException if already exists
  /// \post The new node's initial pose estimate is the identity pose.
  void addNode (NodeId id);

  /// \brief Adds an edge
  /// \param from tail of new edge
  /// \param to head of new edge
  /// \param constraint PoseConstraint expressing head pose in frame of tail pose
  /// \return EdgeId of new edge
  /// \throw UnknownNodeIdException if either id is not known
  EdgeId addEdge (NodeId from, NodeId to, const PoseConstraint& constraint);

  /// \brief Like addEdge, but with a given id
  /// \param from tail of new edge
  /// \param to head of new edge
  /// \param constraint PoseConstraint expressing head pose in frame of tail pose
  /// \param id Desired id for new edge
  /// \throw UnknownNodeIdException if either id is not known
  /// \throw DuplicateEdgeIdException if already exists
  void addEdge (NodeId from, NodeId to, const PoseConstraint& constraint, EdgeId id);

  /// \brief Sets the initial pose estimate attached to this node.
  /// \throw UnknownNodeIdException
  void setInitialPoseEstimate (NodeId n, const geometry_msgs::Pose& pose);

  /// \brief Gets the initial pose estimate attached to this node.
  /// \throw UnknownNodeIdException
  /// \retval A reference valid as long as the node exists
  const geometry_msgs::Pose& getInitialPoseEstimate (NodeId n) const;


  /****************************************
   * Basic const ops
   ****************************************/

  /// Gets set of node ids in graph
  set<NodeId> allNodes() const;

  /// Gets set of edge ids in graph
  set<EdgeId> allEdges() const;

  /// \brief Gets the constraint attached to an edge.  
  /// \retval A reference that is valid so long as edge is in graph.
  /// \throw UnknownEdgeIdException
  const PoseConstraint& getConstraint (EdgeId id) const;

  /// \brief Gets length of an edge
  /// \throw UnknownEdgeIdException
  double edgeLength (EdgeId id) const;

  /// \brief gets adjacent edges to a node
  /// \throws UnknownNodeIdException
  set<EdgeId> incidentEdges (NodeId n) const;

  /// \brief gets incident nodes to an edge, in the order (tail, head).
  /// \throws UnknownEdgeIdException
  pair<NodeId, NodeId> incidentNodes (EdgeId e) const;

  /// \brief Gets neighboring nodes of a node.
  /// \throws UnknownNodeIdException
  std::set<NodeId> neighbors (NodeId n) const;

  /// \brief Checks if a node id already exists
  bool nodeIdExists (NodeId id) const;

  /// \brief Checks if an edge id already exists
  bool edgeIdExists (EdgeId id) const;

  /// \brief Get subgraph over a specific set of nodes
  ///
  /// The returned graph consists of the given nodes, with the same ids
  /// and all edges between them.  It does not share any structure with 
  /// the original graph.
  PoseGraph subgraph (const std::set<NodeId>& nodes) const;


  /****************************************
   * Graph search
   ****************************************/

  /// \brief Computes shortest path using Dijkstra's algorithm
  /// \param src Source node id
  /// \param dest Destination node id
  /// \throws UnknownNodeIdException
  ShortestPathResult shortestPath(NodeId src, NodeId dest) const;

  /// \brief Nodes within a given radius of current one
  /// \param src Source node
  /// \param radius
  /// \throws UnknownNodeIdException
  /// \retval Set of nodes, including src, whose shortest path distance from src is <= radius
  set<NodeId> nearbyNodes (NodeId src, double radius) const;


  /****************************************
   * Point clouds attached to nodes
   ****************************************/

  /// Attach a point cloud to a node.  
  /// \throws UnknownNodeIdException
  void attachCloud (NodeId id, const LocalizedCloud& cloud);

  /// \brief Get the point cloud attached to a node.
  /// \retval A reference that's valid as long as the node's in the graph
  /// \throws UnknownNodeIdException
  /// \throws PoseGraphException (if no cloud found for this node)
  const LocalizedCloud& getCloud (NodeId id) const;

  /// Return true iff the node has an attached cloud
  /// \throws UnknownNodeIdException
  bool hasCloud (NodeId id) const;
  
private:


  /******************************
   * Typedefs
   ******************************/

  typedef map<PoseGraphVertex, double> DistanceMap;
  typedef map<PoseGraphVertex, PoseGraphVertex> PredecessorMap;
  typedef pair<DistanceMap, PredecessorMap> DijkstraResult;


  
  /******************************
   * Private methods
   ******************************/

  // Initialization, used in copy constructor and assignment
  void initializeFrom (const PoseGraphImpl& g);

  // Gets vertex corresponding to a node id
  PoseGraphVertex idVertex (NodeId id) const;

  // Gets edge corresponding to an edge id
  PoseGraphEdge idEdge(EdgeId id) const;

  // Wraps call to dijkstra_shortest_paths
  DijkstraResult dijkstra (const PoseGraphVertex& src) const;

  // Checks if a node's distance in a distance map is above a threshold
  bool outsideRadius (NodeId n, double r, const DistanceMap& distances) const;

  /// \pre \a n is one of the nodes incident to edge \a e
  ///\ retval Node The other incident node
  NodeId otherNode (const NodeId n, const EdgeId e) const;

  /******************************
   * Fields
   ******************************/

  // The underlying boost graph 
  PoseGraphAdjacencyList graph_;

  // Id that will be used for the next new node
  NodeId next_node_id_;

  // Id that will be used for the next new edge
  EdgeId next_edge_id_;

  // Map from NodeId to Vertex
  map<NodeId, PoseGraphVertex> vertex_map_;

  // Map from EdgeId to Edge
  map<EdgeId, PoseGraphEdge> edge_map_;

  // Map from EdgeId to PoseConstraint
  PoseConstraintMap constraints_;

  // Map from vertex to point cloud
  map<PoseGraphVertex, LocalizedCloud> clouds_;
};



  

  

  
   

  




} //namespace pose_graph


#endif // POSE_GRAPH_POSE_GRAPH_H
