/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */


/**
 * \file
 *
 * Defines ScanManager
 *
 * \author Bhaskara Marthi
 */

#ifndef POSE_GRAPH_SCAN_MANAGER_H
#define POSE_GRAPH_SCAN_MANAGER_H

#include <pose_graph/constraint_generator.h>
#include <karto_scan_matcher/karto_scan_matcher.h>
#include <ros/assert.h>
#include <boost/shared_ptr.hpp>
#include <boost/bimap.hpp>
#include <boost/optional.hpp>

namespace pose_graph
{

typedef boost::bimap<NodeId, NodeId> NodeSequenceMap;
typedef std::vector<NodeId> Chain;
typedef boost::shared_ptr<karto_scan_matcher::KartoScanMatcher> KartoMatcherPtr;
typedef sensor_msgs::LaserScan::ConstPtr LaserPtr;
typedef std::map<NodeId, LaserPtr> NodeScanMap;
typedef std::set<NodeId> NodeSet;                         
typedef std::vector<Chain> ChainVec;

/// \brief Internally used class that contains sequential scan matching logic
/// When using the ops, you have to use the WithOptimizedPoses object to set the optimized poses
/// before doing anything else
class ScanManager
{
public:

  ScanManager (const geometry_msgs::Pose2D& laser_offset);
  
  /// Stores the node's scan, precomputes the barycenter, and keeps track of the
  /// sequential order of nodes
  void addNode (NodeId n, LaserPtr scan);

  boost::optional<NodeId> lastNode () const;

  NodeId closestNode (const geometry_msgs::Pose2D& pose, const NodeSet& nodes) const;

  geometry_msgs::Point computeBarycenterInBaseFrame (const LaserPtr scan) const;

  NodeSet nearbyNodes (NodeId n, const geometry_msgs::Point& center, const PoseGraph& graph,
        const double max_distance) const;

protected:

  NodeConstraint scanChainConstraint (const geometry_msgs::Pose2D& pose, const Chain& chain) const;

  template <typename NodeContainer>
  std::vector<karto_scan_matcher::ScanWithPose> makeRefScans (const NodeContainer& nodes) const;

  Chain getChain (NodeId n, const NodeSet& nodes, const NodeSet& processed) const;

  karto_scan_matcher::ScanMatchResult scanMatchNodes (KartoMatcherPtr matcher, 
                                                      const geometry_msgs::Pose& init_pose_estimate, 
                                                      const LaserPtr scan, const NodeSet& nodes) const;

  const geometry_msgs::Pose2D laser_offset_;

  /// \todo this should probably be private
  NodeSequenceMap node_sequence_;

  boost::optional<geometry_msgs::Point> nodeBarycenter (const NodeId n) const;


private:

  NodeScanMap scans_;
  std::map<NodeId, geometry_msgs::Point> barycenters_; // In the base frame
  boost::optional<NodeId> last_node_;
  mutable boost::optional<NodePoseMap> optimized_poses_; // Temporary storage of optimized poses


  geometry_msgs::Point nodeBarycenter (const NodeId n, const geometry_msgs::Pose& pose) const;

  NodeId closestNode (const geometry_msgs::Pose2D& pose, const Chain& chain) const;

  void setOptimizedPoses (const NodePoseMap& optimized_poses) const;
  void unsetOptimizedPoses () const;

  friend class WithOptimizedPoses;


};


/// \brief RAII object for making ScanManager use a given set of optimized poses
struct WithOptimizedPoses
{
  WithOptimizedPoses (const ScanManager* matcher, const NodePoseMap& poses);
  ~WithOptimizedPoses ();
  const ScanManager* matcher;
};


// Utility function for walking along a chain
// Templated so it can take either side of a bimap
template <typename Map>
Chain onewayChain (const Map& m, const NodeId seed, const NodeSet& allowed, const NodeSet& forbidden)
{
  Chain chain;
  NodeId current=seed;
  while (true) {
    if (allowed.find(current)==allowed.end() ||
        forbidden.find(current)!=forbidden.end())
      break;
    chain.push_back(current);
    typename Map::const_iterator pos=m.find(current);
    if (pos==m.end())
      break;
    current = pos->second;
  }
  return chain;
}



} // namespace pose_graph

#endif // include guard
