/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */


/**
 * \file
 *
 * Contains internally used convenience operations involving conversion between
 * ROS, Eigen, and TF, as well as operations on transforms and poses
 *
 * \author Bhaskara Marthi
 */


#ifndef POSE_GRAPH_TRANSFORM_H
#define POSE_GRAPH_TRANSFORM_H

#include <Eigen/Dense>
#include <tf/transform_datatypes.h>
#include <geometry_msgs/Pose.h>
#include <geometry_msgs/Point32.h>
#include <geometry_msgs/Pose2D.h>
#include <nav_msgs/OccupancyGrid.h>

namespace pose_graph
{

using Eigen::Transform3d;
using Eigen::Vector3d;
using Eigen::Quaterniond;
using geometry_msgs::Pose;
using geometry_msgs::Point32;
using geometry_msgs::Pose2D;
using geometry_msgs::Point;


/// Return transformation from pose frame to world frame
Transform3d poseToWorldTransform (const Pose& pose);

/// Return transformation from frame of \a pose1 to frame of \a pose2
Transform3d relativeTransform (const Pose& pose1, const Pose& pose2);

/// Return transformation from frame of \a pose1 to frame of \a pose2
Transform3d relativeTransform (const Pose2D& pose1, const Pose2D& pose2);

/// \a pose1 and \a pose2 are poses specified in some reference frame.
/// This function returns the representation of \a pose1 in the frame at \a pose2.
Pose relativePose (const Pose& pose1, const Pose& pose2);

/// Given point coordinates w.r.t a pose, return its coordinates in reference frame
Point pointInReferenceFrame (const Point& p, const Pose& pose);

/// Return transformation that sends \a pose1 to \a pose2.
/// As opposed to relativeTransform which maps coordinates w.r.t pose1 to coordinates w.r.t pose2
Transform3d transformBetween (const Pose& pose1, const Pose& pose2);

/// Apply an Eigen transformation to a ROS pose
Pose applyTransform (const Transform3d& trans, const Pose& pose);

/// Apply an Eigen transformation to a ROS point
Point32 transformPoint (const Transform3d& trans, const Point32& pt);

/// Apply an Eigen transformation to a ROS point
Point transformPoint (const Transform3d& trans, const Point& pt);

/// Convert pose into x, y, yaw
Vector3d poseToVec (const Pose& p);

/// Perturb x, y components of pose by uniform noise between -r and r
Pose perturbPose (const Pose& pose, double r);

/// Convert ROS quaternion to Eigen Vector4d in x,y,z,w order
Eigen::Vector4d quaternionMsgToVector4d (const geometry_msgs::Quaternion& m);

/// Convert a ROS point to an Eigen 4d vector with 1 at the end
Eigen::Vector4d translationToVector4d(const geometry_msgs::Point& p);

/// Project Pose to Pose2D
Pose2D projectToPose2D (const Pose& pose);

/// Convert to Pose
Pose convertToPose (const Pose2D& pose2d);

/// Construct a pose2d
Pose2D makePose2D (double x, double y, double theta=0.0);

/// Transform an occupancy grid (just transforms the origin)
nav_msgs::OccupancyGrid transformGrid (const Transform3d& trans, const nav_msgs::OccupancyGrid& g);

/// Return dimensions of a map centered at pose of the given size (meters) and resolution
nav_msgs::MapMetaData getMapInfo (const Pose& pose, const double size, const double resolution);


} // namespace pose_graph

#endif
