/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */


/**
 * \file
 *
 * Implementation of FakeLoopCLosureGenerator
 *
 * \author Bhaskara Marthi
 */

#include <pose_graph/constraints/fake_loop_closure.h>
#include <pose_graph/utils.h>
#include <pose_graph/transforms.h>
#include <boost/foreach.hpp>

namespace pose_graph
{

namespace gm=geometry_msgs;

FakeLoopClosureGenerator::FakeLoopClosureGenerator (const double proximity_threshold) :
  fake_loc_sub_(nh_.subscribe("fake_localization", 1, &FakeLoopClosureGenerator::locCB, this)),
  proximity_threshold_(proximity_threshold)
{}



NodeConstraintVector FakeLoopClosureGenerator::getConstraints (const PoseGraph& graph, 
                                                               const NodeId new_node, 
                                                               const NodePoseMap& optimized_poses)
{
  NodeConstraintVector constraints;
  if (haveRecentlyReceivedPose()) {
    const gm::Pose& pose = last_pose_->pose;
    BOOST_FOREACH (const NodePoseMap::value_type entry, true_poses_) {
      const double dist = euclideanDistance(entry.second.position, pose.position);
      if (dist<proximity_threshold_) {
        PoseConstraint constraint = makeConstraint(relativeTransform(pose, entry.second));
        for (unsigned i=0; i<6; i++) 
          constraint.precision(i,i)=10.0;
        constraints.push_back(NodeConstraint(entry.first, constraint));
        ROS_DEBUG_STREAM_NAMED ("fake_place_rec", "Adding constraint from " << entry.first << " to "
                                << new_node << " : " << constraint);
      }
    }
    true_poses_[new_node] = last_pose_->pose;
  }
  else {
    ROS_DEBUG_NAMED ("fake_loop_closure", "Not adding constraint as no recent fake_localization received");
  }
  return constraints;
}


void FakeLoopClosureGenerator::locCB (const gm::PoseWithCovarianceStamped& msg)
{
  gm::PoseStamped pose;
  pose.header = msg.header;
  pose.pose = msg.pose.pose;
  last_pose_ = pose;
}

bool FakeLoopClosureGenerator::haveRecentlyReceivedPose () const
{
  const ros::Time cutoff = ros::Time::now() - ros::Duration(LOCALIZATION_RECENCY_THRESHOLD);
  return (last_pose_ && last_pose_->header.stamp > cutoff);
}



} // namespace pose_graph


