/**
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */


/**
 * \file
 *
 * Implementation of geometry.h
 */

#include <pose_graph/geometry.h>
#include <tf/transform_datatypes.h>
#include <boost/foreach.hpp>

namespace pose_graph 
{

namespace sm=sensor_msgs;
namespace gm=geometry_msgs;

using Eigen::Quaterniond;
using Eigen::Vector3d;
using Eigen::Transform3d;
using geometry_msgs::Pose;
using geometry_msgs::Point;
using geometry_msgs::Point32;

PoseConstraint::PoseConstraint () : translation(Vector3d(0,0,0)), rotation(Quaterniond(0,0,0,1))
{
}

PoseConstraint::PoseConstraint (const Vector3d& translation, const Quaterniond& rotation)
  : translation(translation), rotation(rotation)
{
}

PoseConstraint::PoseConstraint (const Eigen::Vector3d& translation, const Eigen::Quaterniond& rotation, 
                                const Eigen::Matrix<double, 6, 6>& precision) :
  translation(translation), rotation(rotation), precision(precision)
{
}



// Print a constraint
ostream& operator<< (ostream& str, const PoseConstraint& constraint)
{
  tf::Quaternion q(constraint.rotation.x(), constraint.rotation.y(),
                   constraint.rotation.z(), constraint.rotation.w());
  str << "[" << constraint.translation.x() << ", " << constraint.translation.y() 
      << ", " << tf::getYaw(q) << "]";
  return str;
}


// Length of a vector
double length (const Vector3d& v)
{
  return sqrt(v.dot(v));
}

// Length of translation part of pose constraint
double length (const PoseConstraint& constraint)
{
  return length(constraint.translation);
}



// Given a pose constraint, return the expected relative pose
Pose getPose (const PoseConstraint& c)
{
  Pose p;
  p.position.x = c.translation.x();
  p.position.y = c.translation.y();
  p.position.z = c.translation.z();
  p.orientation.x = c.rotation.x();
  p.orientation.y = c.rotation.y();
  p.orientation.z = c.rotation.z();
  p.orientation.w = c.rotation.w();
  return p;
}


// Given a transform between two frames, get a PoseConstraint.  
// We set the precision matrix to the identity.
PoseConstraint makeConstraint (const Transform3d& trans)
{
  PrecisionMatrix precision = Eigen::MatrixXd::Identity(6, 6);
  return PoseConstraint(trans.translation(), Quaterniond(trans.rotation()), precision);
}


double euclideanDistance (const Point& p1, const Point& p2)
{
  return length(Vector3d(p1.x-p2.x, p1.y-p2.y, p1.z-p2.z));
}


Point convertToPoint (const Point32& p)
{
  Point p2;
  p2.x = p.x;
  p2.y = p.y;
  p2.z = p.z;
  return p2;
}


gm::Point computeBarycenter (const sm::LaserScan& scan)
{
  gm::Point sum;
  double angle=scan.angle_min;
  unsigned num_hits=0;

  /// Sum up the hit points
  BOOST_FOREACH (const double range, scan.ranges) {
    ROS_ASSERT (angle>=scan.angle_min && angle<=scan.angle_max+scan.angle_increment/2);
    if (range <= scan.range_max) {
      sum.x += range*cos(angle);
      sum.y += range*sin(angle);
      num_hits++;
    }
    angle += scan.angle_increment;
  }
  ROS_ASSERT_MSG (num_hits>0, "Can't find barycenter of scan: no in-range hits out of %zu", scan.ranges.size());

  gm::Point result;
  result.x = sum.x/num_hits;
  result.y = sum.y/num_hits;
  return result;
}
  

double euclideanDistance (const gm::Pose2D& p1, const gm::Pose2D& p2)
{
  const double dx = p1.x - p2.x;
  const double dy = p1.y - p2.y;

  return sqrt(dx*dx + dy*dy);
}

} // namespace
