/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/**
 * \file graph_mapper.cpp
 * A ros node that constructs a pose graph over time from scan data and odometry
 */


#include <pose_graph/pose_graph.h>
#include <pose_graph/constraint_generator.h>
#include <pose_graph/utils.h>
#include <pose_graph/scan_match_localization.h>
#include <pose_graph/spa_conversion.h>
#include <pose_graph/transforms.h>
#include <pose_graph/pose_graph_message.h>
#include <pose_graph/Edge.h>
#include <pose_graph/exception.h>
#include <occupancy_grid_utils/ray_tracer.h>
#include <ros/ros.h>
#include <ros/time.h>
#include <tf/transform_datatypes.h>
#include <tf/transform_listener.h>
#include <tf/transform_broadcaster.h>
#include <tf_conversions/tf_eigen.h>
#include <sensor_msgs/LaserScan.h>
#include <pose_graph/LocalizedCloud.h>
#include <pose_graph/GraphLocalization.h>
#include <laser_geometry/laser_geometry.h>
#include <geometry_msgs/PoseStamped.h>
#include <visualization_msgs/MarkerArray.h>
#include <boost/thread.hpp>
#include <boost/foreach.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/optional.hpp>
#include <string>
#include <algorithm>
#include <std_msgs/Int64.h>
#include <std_msgs/UInt64.h>
#include <queue>


namespace pose_graph
{

namespace gu=occupancy_grid_utils;
namespace sm=sensor_msgs;
namespace gm=geometry_msgs;
namespace nm=nav_msgs;
namespace vm=visualization_msgs;

using boost::format;
using XmlRpc::XmlRpcValue;
using std::string;
using std::map;
using Eigen::Vector3d;
using Eigen::Quaterniond;
using Eigen::Transform3d;
using Eigen::Translation3d;
using ros::NodeHandle;
using ros::WallTimerEvent;
using ros::WallDuration;
using ros::Duration;
using ros::Time;
using std_msgs::Int64;
using std_msgs::UInt64;
using boost::optional;

typedef boost::mutex::scoped_lock Lock;
typedef set<NodeId> NodeSet;

/************************************************************
 * Constants
 ************************************************************/

const double DEFAULT_POS_THRESHOLD = 2.0;
const double DEFAULT_ANGLE_THRESHOLD = 1.57;
const double STATUS_CHECK_INT = 5.0;

/************************************************************
 * Class that holds node state
 ************************************************************/

class GraphMapper
{
public:

  /// Constructor initializes node based on ros params, then sets up the various timers and callbacks, then returns
  GraphMapper();

private:

  /****************************************
   * Ops
   ****************************************/

  // Set up the constraint generators from the parameter server
  void setupConstraintGenerators () ;

  // Add a new node if necessary
  void updateGraph (const WallTimerEvent& e);

  // Update the grid node data structures based on last_node_
  // Specifically, grid_node_to_nodes_, node_to_grid_node_, and current_grid_node_
  void updateGrid ();

  // Update the fixed frame pose
  void updateFixedFramePose ();

  // Actually add a node to graph and attach last scan
  void addNewNode ();

  // Check if two poses are far enough apart to be separate nodes
  bool sufficientlyFarApart (const gm::Pose& p1, const gm::Pose& p2) const;

  // Save the last laser scan
  void scanCallback (const sm::LaserScan::ConstPtr& scan);

  // Visualize grid containing a particular node
  void visualizeGridCallback (const UInt64& n);

  // Figure out transformation used in visualization
  Transform3d getVisualizationTransform () const;

  // Publish visualization
  void visualize (const WallTimerEvent& e);

  // Use graph optimizer to improve pose estimates
  void optimizePoses (const WallTimerEvent& e);

  // Publish the current graph
  void publishGraph (const WallTimerEvent& e);

  // Get the most recent fixed frame pose.
  gm::PoseStamped currentFixedFramePose () const;

  // Current local grid at node \a n (in fixed frame)
  nm::OccupancyGrid localGrid (const NodeId n);

  // Get local cloud for point cloud at node \a n
  gu::LocalizedCloud makeLocalizedCloud (const NodeId n);
  
  // Periodically check that the node state looks ok
  void statusCheck (const WallTimerEvent& e);

  // Update the localization estimate
  void updateLocalization (const WallTimerEvent& e);

  // Figure out which grid node corresponds to a given node
  optional<NodeId> getGridNode (const NodeId n) const;  

  /****************************************
   * Associated objects
   ****************************************/

  mutable boost::mutex mutex_; 
  NodeHandle nh_;
  ros::WallTimer update_timer_, visualization_timer_, optimization_timer_, graph_publication_timer_, 
    status_check_timer_, localization_timer_;
  ros::Subscriber scan_sub_, visualize_grid_sub_;
  ros::Publisher vis_marker_pub_, marker_array_pub_, vis_scan_pub_, graph_pub_,
    current_frame_pub_, diff_pub_, local_grid_pub_;
  tf::TransformListener tf_;
  tf::TransformBroadcaster tf_broadcaster_;
  vector<ConstraintGenPtr> constraint_generators_;
  ScanMatchLocalization localizer_;
  

  /****************************************
   * Local state; access locked by mutex_
   ****************************************/

  // Pointer to last scan
  LaserPtr last_scan_;

  // True if at least one fixed frame pose has been received so far
  bool fixed_frame_pose_valid_;

  // Last known fixed frame pose; valid iff fixed_frame_pose_valid_ is true
  gm::PoseStamped last_fixed_frame_pose_;

  // Maps from node to fixed frame pose at that node
  NodePoseMap node_fixed_frame_poses_;

  // True iff at least one node has been added (and so the other state variables are meaningful)
  bool nodes_added_;

  // Last added node's id; valid iff nodes_added_ is true
  NodeId last_node_;

  // The graph that's built up over time
  PoseGraph graph_;

  // This holds the results of optimization
  // At any point, its keys will be a subset (since new nodes may have been added) of the nodes in graph_
  NodePoseMap optimized_node_poses_;

  // Maps node id to timestamp when we were at that pose
  map<NodeId, Time> node_stamps_;

  // Valid iff nodes_added_ is true
  Time last_node_time_;
  gm::Pose last_node_pose_;

  // Maps each node to the associated grid node
  map<NodeId, NodeId> node_to_grid_node_;
  
  // Maps each grid node to the set of nodes whose clouds are overlaid to form the grid
  map<NodeId, NodeSet> grid_node_to_nodes_;

  // Current grid node (if there is one)
  optional<NodeId> current_grid_node_;

  // Localization state
  optional<LocalizationState> localization_;

  // Queue of unpublished diffs to the graph
  std::queue<PoseGraphDiff> diffs_;

  // Node whose grid we visualize next
  optional<NodeId> node_to_visualize_;

  /****************************************
   * Parameters
   ****************************************/

  // The frame in which scans come in
  const string sensor_frame_;

  // The base frame of the robot
  const string base_frame_;

  // A locally fixed frame (typically, odometry)
  const string fixed_frame_;

  // Min displacement (in meters) before new graph node created
  const double pos_threshold_;

  // Min angular displacement (in radians) before new graph node created
  const double angle_threshold_;

  // Whether to do optimization
  const bool optimize_;

  // The side length in meters of the local grids 
  const double local_grid_size_;

  // The resolution of the local grids
  const double local_grid_resolution_;
  

};


/************************************************************
 * Utilities
 ***********************************************************/

unsigned getUniqueId () 
{
  static unsigned id=0;
  return id++;
}
  

/************************************************************
 * Initialization
 ************************************************************/

template <class T>
T getPrivateParam (const string& name, const T& default_val)
{
  T val;
  NodeHandle nh("~");
  nh.param(name, val, default_val);
  ROS_DEBUG_STREAM_NAMED ("init", "Set param " << name << " to value " 
                          << val << " (default was " << default_val << ")");
  return val;
}


void GraphMapper::setupConstraintGenerators () 
{
  NodeHandle private_nh("~");
  XmlRpcValue constraint_sources;
  if (!private_nh.getParam("constraint_sources", constraint_sources))
    throw NonexistentParamException("~constraint_sources");
  const XmlRpcValue::Type type = constraint_sources.getType();
  if (type != XmlRpcValue::TypeArray)
    throw PoseGraphException(format("~constraint_sources parameter was not an array (type was %1%)") % type);
  for (int i=0; i<constraint_sources.size(); i++) {
    XmlRpcValue source = constraint_sources[i];
    const XmlRpcValue::Type type = source.getType();
    if (type != XmlRpcValue::TypeString)
      throw PoseGraphException(format("~constraint_sources[%1%] was not a string (type was %1%)") % type);

    const NodeHandle source_nh(private_nh, (string) source);
    ROS_DEBUG_STREAM_NAMED ("init", "Initializing constraint " << (string) source);
    constraint_generators_.push_back(createConstraintGenerator(source_nh));
  }
}

  

GraphMapper::GraphMapper () :
  nh_(), update_timer_(nh_.createWallTimer(WallDuration(0.1), &GraphMapper::updateGraph, this)),
  visualization_timer_(nh_.createWallTimer(WallDuration(2.0), &GraphMapper::visualize, this)),
  optimization_timer_(nh_.createWallTimer(WallDuration(1.0), &GraphMapper::optimizePoses, this)),
  graph_publication_timer_(nh_.createWallTimer(WallDuration(1.0), &GraphMapper::publishGraph, this)),
  status_check_timer_(nh_.createWallTimer(WallDuration(STATUS_CHECK_INT), &GraphMapper::statusCheck, this)),
  localization_timer_(nh_.createWallTimer(WallDuration(0.5), &GraphMapper::updateLocalization, this)),
  scan_sub_(nh_.subscribe("scan", 10, &GraphMapper::scanCallback, this)),
  visualize_grid_sub_(nh_.subscribe("visualize_grid", 10, &GraphMapper::visualizeGridCallback, this)),
  vis_marker_pub_(nh_.advertise<vm::Marker> ("visualization_marker", 10)),
  marker_array_pub_(nh_.advertise<vm::MarkerArray> ("visualization_marker_array", 1000)),
  vis_scan_pub_(nh_.advertise<sm::PointCloud> ("pose_graph_scans", 100)),
  graph_pub_(nh_.advertise<PoseGraphMessage> ("graph", 10)),
  current_frame_pub_(nh_.advertise<Int64> ("current_node", 10)),
  diff_pub_(nh_.advertise<PoseGraphDiff> ("graph_diffs", 1000)), // We really don't want this to overflow
  local_grid_pub_(nh_.advertise<nm::OccupancyGrid> ("graph_mapper_local_grid", 10)),
  localizer_(makePose2D(0, 0.275, 0), 5.0),
  fixed_frame_pose_valid_(false), nodes_added_(false), 
  sensor_frame_(getPrivateParam<string>("sensor_frame", "base_laser_link")),
  base_frame_(getPrivateParam<string>("base_frame", "base_link")),
  fixed_frame_(getPrivateParam<string>("fixed_frame", "odom")),
  pos_threshold_(getPrivateParam<double>("position_displacement_threshold", DEFAULT_POS_THRESHOLD)),
  angle_threshold_(getPrivateParam<double>("angle_displacement_threshold", DEFAULT_ANGLE_THRESHOLD)),
  optimize_(getPrivateParam<bool>("optimize", true)),
  local_grid_size_(getPrivateParam("local_grid_size", 20.0)),
  local_grid_resolution_(getPrivateParam("local_grid_resolution", 0.05))
{
  setupConstraintGenerators();
}

/************************************************************
 * Callbacks
 ************************************************************/

// Store the most recent laser scan
void GraphMapper::scanCallback (const LaserPtr& scan)
{
  Lock lock(mutex_);
  ROS_DEBUG_COND_NAMED (!last_scan_, "pose_graph_node", "First scan received");
  last_scan_ = scan;
}

// Visualize the local grid for a particular node
void GraphMapper::visualizeGridCallback (const UInt64& n) 
{
  node_to_visualize_ = n.data;
}



             
/************************************************************
 * Thread 1: periodically add a new node to graph
 * if we're far enough away from the last one, and connect
 * it to previous nodes by calling the constraint generators
 ************************************************************/

void GraphMapper::updateGraph (const WallTimerEvent& e)
{
  updateFixedFramePose();
  Lock lock(mutex_);
  if ((nodes_added_ && 
       sufficientlyFarApart(last_fixed_frame_pose_.pose, keyValue(node_fixed_frame_poses_, last_node_))) ||
      (!nodes_added_ && last_scan_ && fixed_frame_pose_valid_)) {

    addNewNode(); // Sets last_node_

    // Also create the diff object that we'll add to the publication queue
    // As a postcondition of addNewNode, last_node_ is guaranteed to have an attached cloud and initial pose
    PoseGraphDiff diff; 
    ROS_ASSERT (graph_.hasCloud(last_node_));
    diff.cloud_ids.push_back(last_node_);
    diff.clouds.push_back(graph_.getCloud(last_node_));
    Node diff_node;
    diff_node.id = last_node_;
    diff_node.initial_pose_estimate = graph_.getInitialPoseEstimate(last_node_);
    diff.new_nodes.push_back(diff_node);
    
    BOOST_FOREACH (ConstraintGenPtr gen, constraint_generators_) {
      NodeConstraintVector constraints = gen->getConstraints(graph_, last_node_, node_fixed_frame_poses_);
      BOOST_FOREACH (const NodeConstraint& c, constraints) {
        const EdgeId e = graph_.addEdge(c.target, last_node_, c.constraint);
        
        // Update the diff
        Edge diff_edge;
        diff_edge.src = c.target;
        diff_edge.dest = last_node_;
        diff_edge.constraint = constraintToRos(c.constraint);
        diff_edge.id = e;
        diff.new_edges.push_back(diff_edge);
      }
    }
    
    updateGrid();
    diffs_.push(diff);
  }
}

// Update the variable last_fixed_frame_pose_ to hold the
// most recent fixed frame pose, and set fixed_frame_pose_valid_ to true
void GraphMapper::updateFixedFramePose ()
{
  Time now = Time::now();
  if (tf_.waitForTransform(fixed_frame_, base_frame_, now, ros::Duration(1.0))) {
    gm::PoseStamped p;
    p.pose.position.x = p.pose.position.y = p.pose.position.z = 0.0;
    p.pose.orientation = tf::createQuaternionMsgFromYaw(0.0);
    p.header.frame_id = base_frame_;
    p.header.stamp = now;
    tf::Stamped<tf::Pose> base_pose, fixed_pose;
    tf::poseStampedMsgToTF(p, base_pose);

    /// \todo this can technically throw a tf exception if the last few lines take > 10 seconds to execute
    tf_.transformPose(fixed_frame_, base_pose, fixed_pose);

    Lock lock(mutex_);
    tf::poseStampedTFToMsg(fixed_pose, last_fixed_frame_pose_);
    fixed_frame_pose_valid_ = true;
    
    ROS_DEBUG_STREAM_NAMED ("pose_graph_update_fixed_frame_pose", "Updated fixed frame pose to "
                            << poseToVec(last_fixed_frame_pose_.pose) << " with stamp "
                            << last_fixed_frame_pose_.header.stamp.toSec());
                                      
  }
  else {
    ROS_WARN_STREAM ("Skipping update of fixed frame pose due to lack of transform from " << 
                     base_frame_ << " to " << fixed_frame_);
  }
}


// Are two poses sufficiently far apart that a new node should be added to the graph?
bool GraphMapper::sufficientlyFarApart (const gm::Pose& p1, const gm::Pose& p2) const
{
  const double dx = p1.position.x - p2.position.x;
  const double dy = p1.position.y - p2.position.y;
  const double dtheta = abs(tf::getYaw(p1.orientation) - tf::getYaw(p2.orientation));
  return (dx*dx + dy*dy > pos_threshold_*pos_threshold_) || (dtheta > angle_threshold_);
}

// Add a new node at the current position in the graph
void GraphMapper::addNewNode ()
{
  ROS_ASSERT (last_scan_ && fixed_frame_pose_valid_);
  ROS_ASSERT_MSG (last_scan_->header.frame_id == sensor_frame_, 
                  "Code currently assumes scans are in default sensor frame");

  /****************************************
   * 1. Add the node itself
   ****************************************/

  NodeId n = graph_.addNode();
  graph_.setInitialPoseEstimate(n, last_fixed_frame_pose_.pose);
  node_fixed_frame_poses_[n] = last_fixed_frame_pose_.pose;
  node_stamps_[n] = last_fixed_frame_pose_.header.stamp;
  ROS_DEBUG_NAMED ("pose_graph_node_update", 
                   "Added node %u with stamp %.2f, at [%.2f, %.2f, %.2f].",
                   n, node_stamps_[n].toSec(), last_fixed_frame_pose_.pose.position.x, 
                   last_fixed_frame_pose_.pose.position.y, 
                   tf::getYaw(last_fixed_frame_pose_.pose.orientation));
  nodes_added_ = true;
  last_node_ = n;
  

  /****************************************
   * 2. Store the projected and transformed
   * laser scan
   ****************************************/

  // We want to store the entire cloud with respect to the current base frame
  // Unfortunately, transformLaserScanToPointCloud assumes the target frame is fixed
  // So we first project to a (more or less) fixed frame, then transform to the 
  // sensor frame at a particular time point
  
  double scan_duration = last_scan_->ranges.size()*last_scan_->time_increment;
  if (!tf_.waitForTransform(fixed_frame_, sensor_frame_, 
                            last_scan_->header.stamp + Duration(scan_duration), Duration(2.0)))
  {
    ROS_FATAL_STREAM ("Timed out waiting for transform from " << sensor_frame_ << " to " << fixed_frame_);
    ROS_BREAK();
  }
  sm::PointCloud fixed_frame_cloud;
  laser_geometry::LaserProjection projector_;
  projector_.transformLaserScanToPointCloud (fixed_frame_, *last_scan_, fixed_frame_cloud, tf_);

  if (!tf_.waitForTransform(sensor_frame_, fixed_frame_, node_stamps_[n], ros::Duration(2.0))) {
    ROS_FATAL_STREAM ("Timed out waiting for transform from " << fixed_frame_ << " to " << base_frame_);
    ROS_BREAK();
  }

  LocalizedCloud sensor_frame_cloud;
  tf_.transformPointCloud (sensor_frame_, node_stamps_[n], fixed_frame_cloud, 
                           fixed_frame_, sensor_frame_cloud.cloud);
  try {
    geometry_msgs::PoseStamped id, sensor_pose;
    id.header.stamp = ros::Time();
    id.header.frame_id = sensor_frame_;
    id.pose.orientation.w = 1.0; // id is now the identity pose
    tf_.transformPose(base_frame_, id, sensor_pose);
    sensor_frame_cloud.sensor_pose = sensor_pose.pose;
  } 
  catch (tf::TransformException& e) {
    ROS_FATAL_STREAM ("Received tf exception '" << e.what() << "' when trying to compute sensor pose.  Aborting.");
    ROS_BREAK();
  }
  graph_.attachCloud(n, sensor_frame_cloud);

  /****************************************
   * 3. Set the localization
   ****************************************/
  
  const Eigen::Transform3d node_to_fixed_frame = poseToWorldTransform(node_fixed_frame_poses_[n]);
  localizer_.addNode(n, last_scan_);
  gm::Pose identity;
  identity.orientation.w = 1.0;
  localization_=LocalizationState(n, node_to_fixed_frame, identity);

}


bool withinDistance (const PoseGraph& g, const NodeId n1, const NodeId n2, const double d)
{
  ShortestPathResult res=g.shortestPath(n1, n2);
  return res.path_found && res.cost < d;
}


void GraphMapper::updateGrid ()
{  
  // Start up a new grid if we're too far from the current one
  if (!(current_grid_node_ && 
        withinDistance(graph_, *current_grid_node_, last_node_, local_grid_size_/3))) {
    current_grid_node_ = last_node_;
    grid_node_to_nodes_[last_node_] = NodeSet();
  }

  // Add this node to the current grid
  node_to_grid_node_[last_node_] = *current_grid_node_;
  grid_node_to_nodes_[*current_grid_node_].insert(last_node_);
  ROS_DEBUG_NAMED ("pose_graph_node_update", "Adding %u to grid %u", last_node_, *current_grid_node_);
}


/************************************************************
 * Thread 2: periodically optimize the entire
 * graph using sparse pose adjustment
 ************************************************************/


void GraphMapper::optimizePoses (const WallTimerEvent& e)
{
  Lock lock(mutex_);
  optimized_node_poses_ = 
    optimize_ ? 
    optimizeGraph(graph_) :
    node_fixed_frame_poses_;
}



/************************************************************
 * Thread 3: publish visualization
 ************************************************************/


// Say you have a node at pose p and an outgoing edge labelled with constraint c.
// This returns the expected position of the node on the other side of the edge.
gm::Pose applyConstraint (const gm::Pose& p, const PoseConstraint& c)
{
  const Transform3d trans = poseToWorldTransform(p);
  const gm::Pose rel_pose = getPose(c);
  return applyTransform(trans, rel_pose);
}

// Get the transform from the frame of optimized_node_poses_ to the fixed frame
Transform3d GraphMapper::getVisualizationTransform () const
{
  ROS_ASSERT (optimized_node_poses_.size() > 0);
  const NodeId n = optimized_node_poses_.rbegin()->first;
  const gm::Pose opt_pose = keyValue(optimized_node_poses_, n);
  const gm::Pose fixed_frame_pose = keyValue(node_fixed_frame_poses_, n);
  const Transform3d trans = transformBetween(opt_pose, fixed_frame_pose);

  ROS_DEBUG_STREAM_NAMED ("pose_graph_node_vis", "Opt pose is " << poseToVec(opt_pose) << 
                          ", fixed frame pose is " << poseToVec(fixed_frame_pose) << 
                          ", and transform is " << trans.matrix());
  
  return trans;
}

void addTransformedPoints (const LocalizedCloud& cloud, const gm::Pose& pose, sm::PointCloud* cloud_out)
{
  const Transform3d base_to_world = poseToWorldTransform(pose);
  const Transform3d sensor_to_base = poseToWorldTransform(cloud.sensor_pose);
  const Transform3d sensor_to_world = base_to_world*sensor_to_base;

  BOOST_FOREACH (const gm::Point32 pt, cloud.cloud.points) 
    cloud_out->points.push_back(transformPoint(sensor_to_world, pt));
}


// Visualize graph and scans
void GraphMapper::visualize (const WallTimerEvent& e) 
{
  const unsigned NODE_MARKER_ID_BASE = 1000;
  
  Lock lock(mutex_);
  vm::Marker edges, loc;
  vm::Marker node_prototype;
  vm::MarkerArray nodes;
  sm::PointCloud cloud;

  unsigned next_ind=0;
  edges.header.frame_id = node_prototype.header.frame_id = loc.header.frame_id = fixed_frame_;
  node_prototype.id = -42;
  edges.id = next_ind++;
  loc.id = next_ind++;
  node_prototype.header.stamp = edges.header.stamp = loc.header.stamp = Time::now();
  edges.ns = node_prototype.ns = loc.ns = "pose_graph";
  node_prototype.type = vm::Marker::SPHERE;
  edges.type = vm::Marker::LINE_LIST;
  loc.type = vm::Marker::ARROW;
  edges.action = node_prototype.action = loc.action = vm::Marker::ADD;
  edges.scale.x = 0.1;
  node_prototype.scale.x = 0.2;
  node_prototype.scale.y = 0.2;
  node_prototype.scale.z = 0.2;
  node_prototype.pose.orientation.w = 1.0;
  loc.scale.x = 0.2;
  loc.scale.y = 0.4;
  edges.color.b = node_prototype.color.r = 1.0;
  loc.color.r = .64;
  loc.color.g = .1;
  loc.color.b = .9;
  edges.color.a = loc.color.a = node_prototype.color.a = 1.0;



  if (optimized_node_poses_.size() > 0) {

    const Transform3d trans = getVisualizationTransform();

    BOOST_FOREACH (const NodePoseMap::value_type& entry, optimized_node_poses_) {

      // Publish sphere at node position
      const NodeId n = entry.first;
      const gm::Pose pose = applyTransform(trans, entry.second);
      vm::Marker node = node_prototype;
      const size_t node_ind = nodes.markers.size();
      nodes.markers.resize(node_ind+1);
      nodes.markers[node_ind] = node_prototype;
      nodes.markers[node_ind].pose.position = pose.position;
      nodes.markers[node_ind].id = n + NODE_MARKER_ID_BASE;

      // Add scan at node to point cloud
      if (graph_.hasCloud(n))
        addTransformedPoints(graph_.getCloud(n), pose, &cloud);
      // This loops over the constraints, and for each one that is rooted at the current node, draws a line
      BOOST_FOREACH (const EdgeId e, graph_.incidentEdges(n)) {
        if (graph_.incidentNodes(e).first == n) {
          edges.points.push_back(pose.position);
          const gm::Pose pose2 = applyConstraint(entry.second, graph_.getConstraint(e));
          edges.points.push_back(applyTransform(trans, pose2).position);
        }
      }
    }

    /* Commenting this out because the grid visualization really belongs in navigation graph
    if (contains(optimized_node_poses_, *current_grid_node_))
      current_local_grid_pub_.publish(transformGrid(trans, localGrid(*current_grid_node_)));
    */

    if (node_to_visualize_) {
      if (contains(optimized_node_poses_, *node_to_visualize_)) {
        optional<NodeId> grid_node = getGridNode(*node_to_visualize_);
        if (grid_node)
          local_grid_pub_.publish(transformGrid(trans, localGrid(*grid_node)));
        else
          ROS_WARN ("No attached grid for node %u", *node_to_visualize_);
      }
      else
        ROS_WARN ("Can't visualize grid for %u as we don't have an optimized pose for it", *node_to_visualize_);
    }

    // Localization
    if (localization_) {
      NodePoseMap::const_iterator iter=optimized_node_poses_.find(localization_->ref_node);
      if (iter!=optimized_node_poses_.end()) {
        loc.points.push_back(applyTransform(trans, iter->second).position);
        Eigen::Transform3d node_to_vis=trans*localization_->node_to_fixed_frame;
        loc.points.push_back(applyTransform(node_to_vis, localization_->node_frame_pose).position);
      }
    }
    else
      loc.action = vm::Marker::DELETE;
  }

  marker_array_pub_.publish(nodes);
  vis_marker_pub_.publish(edges);
  vis_marker_pub_.publish(loc);

  cloud.header.frame_id = fixed_frame_;
  cloud.header.stamp = ros::Time::now();
  vis_scan_pub_.publish(cloud);

}

optional<NodeId> GraphMapper::getGridNode (const NodeId n) const
{
  optional<NodeId> g;
  typedef map<NodeId, set<NodeId> > GridMap;
  BOOST_FOREACH (const GridMap::value_type& e, grid_node_to_nodes_) {
    if (contains(e.second, n)) {
      g = e.first;
      break;
    }
  }
  return g;
}


nm::OccupancyGrid GraphMapper::localGrid (const NodeId n)
{
  const Transform3d node_ref_transform = poseToWorldTransform(keyValue(optimized_node_poses_, n));
  
  gm::Pose local_origin;
  local_origin.position.x = -local_grid_size_/2;
  local_origin.position.y = -local_grid_size_/2;
  local_origin.orientation.w = 1.0;

  nav_msgs::MapMetaData info;
  info.resolution = local_grid_resolution_;
  info.width = local_grid_size_/local_grid_resolution_;
  info.height = local_grid_size_/local_grid_resolution_;
  info.origin = applyTransform(node_ref_transform, local_origin);

  ROS_ASSERT_MSG (contains(grid_node_to_nodes_, n), "Node %u has no associated grid", n);
  vector<gu::LocalizedCloud> clouds;
  BOOST_FOREACH (const NodeId node, grid_node_to_nodes_[n]) {
    if (contains(optimized_node_poses_, node)) {
      clouds.push_back(makeLocalizedCloud(node));
      ROS_DEBUG_NAMED ("local_grid", "Added cloud for node %u", node);
    }
    else 
      ROS_DEBUG_NAMED("local_grid", "Not including %u as its optimized pose is not yet known", node);
  }

  nm::OccupancyGrid grid = gu::overlayClouds(info, clouds);
  grid.header.stamp = ros::Time::now();
  grid.header.frame_id = fixed_frame_;

  return grid;
}




gu::LocalizedCloud GraphMapper::makeLocalizedCloud (const NodeId n)
{
  ROS_ASSERT(contains(optimized_node_poses_, n));
  ROS_ASSERT(graph_.hasCloud(n));
  gu::LocalizedCloud fixed_frame_cloud;
  const LocalizedCloud& base_frame_cloud = graph_.getCloud(n);

  Transform3d pose_to_world = poseToWorldTransform(optimized_node_poses_[n]);
  fixed_frame_cloud.sensor_pose = applyTransform(pose_to_world, base_frame_cloud.sensor_pose);
  transform(base_frame_cloud.cloud.points.begin(), base_frame_cloud.cloud.points.end(), 
            back_inserter(fixed_frame_cloud.points), convertToPoint);

  return fixed_frame_cloud;
}


/************************************************************
 * Thread 4: publish graph and localization w.r.t it
 ************************************************************/

void GraphMapper::publishGraph (const WallTimerEvent& e)
{
  Lock lock(mutex_);
  graph_pub_.publish(poseGraphToRos(graph_));

  // Publish diffs
  while (!diffs_.empty()) {
    diff_pub_.publish(diffs_.front());
    diffs_.pop();
  }

  if (localization_) {

    // Publish the identity of the current node
    Int64 frame_msg;
    frame_msg.data = localization_->ref_node;
    current_frame_pub_.publish(frame_msg);

    // Publish transform to that node's frame to tf
    tf::Transform trans;
    tf::TransformEigenToTF(localization_->node_to_fixed_frame, trans);
    const string node_frame = PoseGraph::nodeFrameName(localization_->ref_node);
    tf::StampedTransform trans_stamped(trans, ros::Time::now(), fixed_frame_, node_frame);
    tf_broadcaster_.sendTransform(trans_stamped);    
  }
}


/************************************************************
 * Thread 5: Occasionally check that everything looks ok
 ************************************************************/

void GraphMapper::statusCheck (const WallTimerEvent& e)
{
  static unsigned count=0;
  if (count++) {
    if (!nodes_added_ || !localization_) {
      ROS_WARN_COND_NAMED (!nodes_added_, "graph_mapper_status_check", "No nodes added yet.");
      ROS_WARN_COND_NAMED (!localization_ && nodes_added_, "graph_mapper_status_check", "No localization");
      ROS_WARN_STREAM_NAMED ("graph_mapper_status_check", "scan_received = " << (last_scan_.get()!=0)
                             << ", fixed_frame_pose_valid_ = " << fixed_frame_pose_valid_);
    }
  }
}


/************************************************************
 * Thread 6: Localization
 ************************************************************/

void GraphMapper::updateLocalization (const WallTimerEvent& e)
{
  Lock lock(mutex_);
  if (localization_ && graph_.allNodes().size() == optimized_node_poses_.size()) {
    // This is a hacky way to get around the fact that not all nodes may be optimized
    // at any given point, but the localization assumes we have an optimized pose for everything
    localization_ = localizer_.nextState(*localization_, last_fixed_frame_pose_.pose,
                                         last_scan_, graph_, optimized_node_poses_);
  }
}

} // namespace pose_graph





int main (int argc, char** argv)
{
  ros::init(argc, argv, "graph_mapper");
  ros::AsyncSpinner spinner(3);
  spinner.start();
  pose_graph::GraphMapper node;
  ROS_INFO ("Initialization complete");
  ros::waitForShutdown();
  return 0;
}
