/* Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */


/**
 * \file
 *
 * Implementation of KartoLaserConstraints
 *
 * \author Bhaskara Marthi
 */

#include <pose_graph/constraints/karto_laser_constraints.h>
#include <pose_graph/transforms.h>
#include <pose_graph/utils.h>
#include <pose_graph/geometry.h>
#include <boost/foreach.hpp>
#include <boost/bind.hpp>
#include <boost/ref.hpp>
#include <queue>

namespace pose_graph
{

namespace ksm=karto_scan_matcher;
namespace sm=sensor_msgs;
namespace gm=geometry_msgs;

using std::vector;
using std::set;
using boost::bind;
using boost::ref;

typedef boost::mutex::scoped_lock Lock;
typedef set<NodeId> NodeSet;
typedef vector<Chain> ChainVec;

KartoLaserConstraints::KartoLaserConstraints (const double max_seq_link_length, 
                                              const unsigned min_seq_chain_length, const double min_seq_response,
                                              const double max_loop_link_length, const unsigned min_loop_chain_length,
                                              const double min_loop_response) :
  max_sequential_link_length_(max_seq_link_length), min_sequential_chain_length_(min_seq_chain_length),
  min_sequential_link_response_(min_seq_response), max_loop_link_length_(max_loop_link_length),
  min_loop_chain_length_(min_loop_chain_length), min_loop_closure_response_(min_loop_response),
  scan_sub_(nh_.subscribe("base_scan", 10, &KartoLaserConstraints::scanCallback, this))
{
}

// Get the laser's pose, relative to base.
gm::Pose2D getLaserOffset (const tf::TransformListener& tf, const sm::LaserScan& scan)
{
  tf::Stamped<tf::Pose> ident;
  tf::Stamped<btTransform> laser_to_base;
  ident.setIdentity();
  ident.frame_id_ = scan.header.frame_id;
  ident.stamp_ = ros::Time();
  
  if (!tf.waitForTransform("base_link", ident.frame_id_, ros::Time(), ros::Duration(5.0))) {
    ROS_FATAL_STREAM ("Didn't get a transform between base_link and " << ident.frame_id_ << " in time");
    ROS_BREAK();
  }
  
  tf.transformPose("base_link", ident, laser_to_base);

  gm::Pose2D laser_pose;
  laser_pose.x = laser_to_base.getOrigin().x();
  laser_pose.y = laser_to_base.getOrigin().y();
  laser_pose.theta = tf::getYaw(laser_to_base.getRotation());
  return laser_pose;
}


// Store the last scan; initialize the ScanManager if necessary
void KartoLaserConstraints::scanCallback (const sm::LaserScan::ConstPtr& scan)
{
  Lock lock(mutex_);
  if (!seq_matcher_) {
    seq_matcher_.reset(new SequentialScanMatcher(getLaserOffset(tf_, *scan), max_sequential_link_length_,
          min_sequential_chain_length_, min_sequential_link_response_));
    ROS_ASSERT (!loop_matcher_);
    loop_matcher_.reset(new LoopScanMatcher(getLaserOffset(tf_, *scan), max_loop_link_length_, min_loop_chain_length_, min_loop_closure_response_));
  }
  last_scan_ = scan;
}


bool targetIs (const NodeConstraint& constraint, const NodeId n)
{
  return (constraint.target == n);
}

// Externally called method that returns sequential and loop constraints
NodeConstraintVector KartoLaserConstraints::getConstraints (const PoseGraph& graph, const NodeId new_node,
                                                    const NodePoseMap& optimized_poses)
{
  NodeConstraintVector constraints = scanOdomConstraints(graph, new_node, optimized_poses);
  const NodeConstraintVector loop_constraints = loopConstraints(graph, new_node, optimized_poses);
  constraints.insert(constraints.end(), loop_constraints.begin(), loop_constraints.end());
  if (last_scan_) {
    seq_matcher_->addNode(new_node, last_scan_);
    loop_matcher_->addNode(new_node, last_scan_);
  }
  return constraints;
}


NodeConstraintVector KartoLaserConstraints::loopConstraints (const PoseGraph& graph, const NodeId new_node,
                                                             const NodePoseMap& optimized_poses)
{
  const gm::Pose current_pose=graph.getInitialPoseEstimate(new_node);
  Lock lock(mutex_);
  if (last_scan_) {
    const boost::optional<NodeId> last_node = seq_matcher_->lastNode();
    if (last_node) {
      ROS_DEBUG_NAMED ("loop_closure", "About to call loop scan matcher");
      ROS_ASSERT(loop_matcher_);

      WithOptimizedPoses opt(loop_matcher_.get(), optimized_poses);
      return loop_matcher_->getConstraints(graph, *last_node, current_pose, last_scan_);
    }
    else 
      ROS_DEBUG_NAMED ("loop_closure", "Not attempting loop closure as this is the first node");
  }
  else
    ROS_DEBUG_NAMED ("loop closure", "Not attempting loop closure as no scans received yet");
    
  return NodeConstraintVector();
}

NodeConstraintVector KartoLaserConstraints::scanOdomConstraints (const PoseGraph& graph, const NodeId new_node,
                                                                 const NodePoseMap& optimized_poses)
{
  const gm::Pose current_pose=graph.getInitialPoseEstimate(new_node);
  Lock lock(mutex_);
  NodeConstraintVector constraints;

  if (!last_scan_) 
    ROS_DEBUG_NAMED ("scan_odom", "No last scan, so not adding any constraints");
  
  else {
    const boost::optional<NodeId> last_node = seq_matcher_->lastNode();
    if (last_node) {
      ROS_DEBUG_NAMED ("scan_odom", "About to call sequential scan matcher");
      ROS_ASSERT (seq_matcher_);
      
      WithOptimizedPoses opt(seq_matcher_.get(), optimized_poses);
      constraints = seq_matcher_->getConstraints(graph, *last_node, current_pose, last_scan_);

      // Add an odom constraint if there's no other constraint linking this node to the previous node
      if (find_if(constraints.begin(), constraints.end(), bind(targetIs, _1, *last_node)) 
          == constraints.end()) {
        ROS_DEBUG_NAMED ("scan_odom", "Adding an extra odom constraint from %u to %u", *last_node, new_node);
        const gm::Pose last_pose = graph.getInitialPoseEstimate(*last_node);
        const PoseConstraint odom_constraint = makeConstraint(relativeTransform(current_pose, last_pose));
        constraints.push_back(NodeConstraint(*last_node, odom_constraint));
      }
    }
    
    else
      ROS_DEBUG_NAMED ("scan_odom", "No constraints since this is the first node");
  }

  return constraints;
}

} // namespace pose_graph
