/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <pose_graph/loop_scan_matcher.h>
#include <pose_graph/geometry.h>
#include <pose_graph/utils.h>
#include <pose_graph/transforms.h>
#include <boost/foreach.hpp>
#include <ros/ros.h>

namespace pose_graph
{

namespace gm=geometry_msgs;
namespace ksm=karto_scan_matcher;
using std::vector;

LoopScanMatcher::LoopScanMatcher (const gm::Pose2D& laser_offset, const double max_link_length,
                                  const unsigned min_chain_length, const double min_response) :
  ScanManager(laser_offset), max_link_length_(max_link_length),
  min_chain_length_(min_chain_length), min_response_(min_response)
{
  ROS_DEBUG_STREAM_NAMED ("init", "Initializing loop scan matcher with max link length " <<
                          max_link_length_ << ", min chain length " << min_chain_length_ <<
                          ", min response " << min_response_);
}


NodeConstraintVector LoopScanMatcher::getConstraints (const PoseGraph& graph, const NodeId ref_node,
                                                      const Pose& init_pose_estimate, const LaserPtr scan) const
{
  NodeConstraintVector constraints;

  // Initialize the matcher based on scan parameters if necessary
  if (!matcher_) {
    ksm::DoubleVector resolutions, sizes;
    
    sizes.push_back(5.0);
    resolutions.push_back(0.05);

    sizes.push_back(0.3);
    resolutions.push_back(0.01);
    
    matcher_.reset(new ksm::KartoScanMatcher(*scan, laser_offset_, sizes, resolutions));
  }
    

  ChainVec candidates = possibleLoopClosures(graph, ref_node, init_pose_estimate, scan);

  BOOST_FOREACH (const Chain& chain, candidates) {
    NodeSet nodes(chain.begin(), chain.end());
    ksm::ScanMatchResult result = scanMatchNodes(matcher_, init_pose_estimate, scan, nodes);
    ROS_DEBUG_STREAM_NAMED ("loop_closure", "Loop closure response against chain " <<
                            toString(chain) << " was " << result.response);
    if (result.response >= min_response_)
      constraints.push_back(scanChainConstraint(result.pose, chain));
  }

  return constraints;
}

// Either n is further than distance from pose, or it doesn't have an optimized pose
bool LoopScanMatcher::furtherThan (const double max_distance, const Point& p, const NodeId n) const
{
  boost::optional<gm::Point> p2 = nodeBarycenter(n);
  return (!p2 || euclideanDistance(p, *p2) > max_distance);
}

ChainVec LoopScanMatcher::possibleLoopClosures (const PoseGraph& graph, const NodeId ref_node,
                                                const Pose& pose, const LaserPtr scan) const
{
  ChainVec chains;
  const gm::Point bary = pointInReferenceFrame(computeBarycenterInBaseFrame(scan), pose);
  const NodeSet nearby = nearbyNodes(ref_node, bary, graph, max_link_length_);

  const NodeSet all_nodes = graph.allNodes();
  NodeSet nodes;
  remove_copy_if(all_nodes.begin(), all_nodes.end(), inserter(nodes, nodes.begin()),
                 bind(&LoopScanMatcher::furtherThan, this, max_link_length_, bary, _1));

  ROS_DEBUG_STREAM_NAMED ("loop_closure", "Looking for possible loop closures.  Barycenter is " << 
                          toString(bary) << " and nearby nodes are " << toString(nearby) <<
                          " and candidate nodes are " << toString(nodes));

  NodeId current = ref_node;
  while (true) {
    const Chain chain = onewayChain(node_sequence_.right, current, nodes, nearby);
    if (chain.size() >= min_chain_length_) {
      ROS_DEBUG_STREAM_NAMED ("loop_closure", "Possible loop closure chain: " << toString(chain));
      chains.push_back(chain);
    }
    if (chain.size() > 0)
      current = *(--chain.end());

    // Continue along sequence/test that we've reached beginning
    NodeSequenceMap::right_map::const_iterator pos = node_sequence_.right.find(current);
    if (pos==node_sequence_.right.end())
      break;
    else
      current = pos->second;
  } 

  return chains;
}



} // namespace pose_graph
