/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/**
 * \file
 * 
 * Implements ops in pose_graph_message.h
 */


#include <pose_graph/pose_graph_message.h>
#include <pose_graph/exception.h>
#include <pose_graph/utils.h>
#include <boost/foreach.hpp>
#include <boost/tuple/tuple.hpp>
#include <boost/thread.hpp>
#include <ros/time.h>

#define foreach BOOST_FOREACH

namespace pose_graph
{

using boost::tie;
using std::string;
using boost::mutex;
using Eigen::Quaterniond;
using Eigen::Vector3d;

const string DUMMY_TOPIC("dummy");

/************************************************************
 * Converting between PoseGraph and PoseGraphMessage
 ************************************************************/

// Uses the assumption that tf::Pose and tf::Transform are the same type
// (will not compile otherwise)
PoseWithPrecision constraintToRos (const PoseConstraint& constraint)
{
  PoseWithPrecision m;
  m.pose.position.x = constraint.translation.x();
  m.pose.position.y = constraint.translation.y();
  m.pose.position.z = constraint.translation.z();

  const Quaterniond::Coefficients& coeffs = constraint.rotation.coeffs();
  m.pose.orientation.x = coeffs.x();
  m.pose.orientation.y = coeffs.y();
  m.pose.orientation.z = coeffs.z();
  m.pose.orientation.w = coeffs.w();

  unsigned ind=0;
  for (unsigned r=0; r<6; r++)
    for (unsigned c=0; c<6; c++) 
      m.precision[ind++] = constraint.precision(r,c);

  return m;
}


PoseGraphMessage poseGraphToRos (const PoseGraph& g)
{
  PoseGraphMessage m;

  foreach (const NodeId n, g.allNodes()) {
    Node node;
    node.id = n;
    node.initial_pose_estimate = g.getInitialPoseEstimate(n);
    m.nodes.push_back(node);
    if (g.hasCloud(n)) {
      m.cloud_ids.push_back(n);
      m.clouds.push_back(g.getCloud(n));
    }
  }

  foreach (const EdgeId e, g.allEdges()) {
    Edge edge;
    edge.id = e;
    tie(edge.src, edge.dest) = g.incidentNodes(e);
    edge.constraint = constraintToRos(g.getConstraint(e));
    m.edges.push_back(edge);
  }
  
  return m;
}


PoseConstraint constraintFromRos (const PoseWithPrecision& m)
{
  PoseConstraint constraint;
  constraint.translation = Vector3d(m.pose.position.x, m.pose.position.y, m.pose.position.z);
  constraint.rotation = Quaterniond(m.pose.orientation.w, m.pose.orientation.x, m.pose.orientation.y,
                                    m.pose.orientation.z);
  unsigned ind=0;
  for (unsigned r=0; r<6; r++)
    for (unsigned c=0; c<6; c++)
      constraint.precision(r,c) = m.precision[ind++];
  return constraint;
}


PoseGraph poseGraphFromRos (const PoseGraphMessage& m)
{
  PoseGraph g;
  
  foreach (const Node& n, m.nodes) {
    g.addNode(n.id);
    g.setInitialPoseEstimate(n.id, n.initial_pose_estimate);
  }

  foreach (const Edge& e, m.edges) 
    g.addEdge(e.src, e.dest, constraintFromRos(e.constraint), e.id);

  for (unsigned i=0; i<m.cloud_ids.size(); i++) 
    g.attachCloud(m.cloud_ids[i], m.clouds[i]);

  return g;
}

/************************************************************
 * Diffs
 ************************************************************/

void applyDiff (PoseGraph* graph, const PoseGraphDiff& msg)
{
  std::set<NodeId> new_nodes;
  BOOST_FOREACH (const Node n, msg.new_nodes) {
    graph->addNode(n.id);
    new_nodes.insert(n.id);
    graph->setInitialPoseEstimate(n.id, n.initial_pose_estimate);
  }
  BOOST_FOREACH (const Edge& e, msg.new_edges)
    graph->addEdge(e.src, e.dest, constraintFromRos(e.constraint), e.id);
  const unsigned num_clouds = msg.cloud_ids.size();
  ROS_ASSERT (msg.clouds.size()==num_clouds);
  for (unsigned i=0; i<num_clouds; i++) {
    ROS_ASSERT(contains(new_nodes, msg.cloud_ids[i]));
    graph->attachCloud(msg.cloud_ids[i], msg.clouds[i]);
  }
}

}
// namespace pose_graph
