/* Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */


/**
 * \file
 *
 * Implementation of ScanManager class
 *
 * \author Bhaskara Marthi
 */
   
#include <pose_graph/scan_manager.h>
#include <pose_graph/geometry.h>
#include <pose_graph/utils.h>
#include <pose_graph/transforms.h>
#include <boost/foreach.hpp>
#include <queue>

namespace pose_graph
{

namespace gm=geometry_msgs;
namespace ksm=karto_scan_matcher;
namespace sm=sensor_msgs;

using std::vector;
using boost::optional;

typedef NodeSequenceMap::value_type SequenceLink;
typedef NodeSequenceMap::left_map::const_iterator LeftIter;
typedef NodeSequenceMap::right_map::const_iterator RightIter;

ScanManager::ScanManager (const gm::Pose2D& offset) : laser_offset_(offset)
{}

void ScanManager::addNode (const NodeId n, const LaserPtr scan)
{
  ROS_ASSERT (!contains(scans_, n) && !contains(barycenters_, n));
  ROS_DEBUG_NAMED ("scan_odom", "Adding node %u", n);
  scans_[n] = scan;

  barycenters_[n] = computeBarycenterInBaseFrame(scan);
  
  if (last_node_) {
    node_sequence_.insert(SequenceLink(*last_node_, n));
    ROS_DEBUG_NAMED ("scan_odom", " added sequence link from %u", *last_node_);
  }
  last_node_ = n;
}
    



optional<NodeId> ScanManager::lastNode () const
{
  return last_node_;
}

ksm::ScanWithPose makeScanWithPose (const NodePoseMap& optimized_poses, const NodeScanMap scans_,
                                    const NodeId n)
{
  return ksm::ScanWithPose(*keyValue(scans_, n), projectToPose2D(keyValue(optimized_poses, n)));
}


template <typename NodeContainer>
vector<ksm::ScanWithPose> ScanManager::makeRefScans (const NodeContainer& nodes) const
{
  vector<ksm::ScanWithPose> loc_scans(nodes.size());
  transform(nodes.begin(), nodes.end(), loc_scans.begin(),
            bind(makeScanWithPose, *optimized_poses_, scans_, _1));
  return loc_scans;
}




Chain ScanManager::getChain (const NodeId seed, const NodeSet& nodes, const NodeSet& processed) const
{
  ROS_ASSERT_MSG (contains(nodes, seed), "Node set does not contain %u", seed);
  ROS_ASSERT_MSG (!contains(processed, seed), "%u was already processed", seed);
  const Chain forward_chain = onewayChain(node_sequence_.left, seed, nodes, processed);
  const Chain backward_chain = onewayChain(node_sequence_.right, seed, nodes, processed);

  Chain chain(forward_chain.size()+backward_chain.size()-1);
  copy(backward_chain.rbegin(), backward_chain.rend(), chain.begin());
  copy(forward_chain.begin()+1, forward_chain.end(), chain.begin() + backward_chain.size());

  return chain;
}

gm::Point ScanManager::computeBarycenterInBaseFrame (const LaserPtr scan) const
{
  const gm::Point offset_center = computeBarycenter(*scan);
  const Transform3d offset_transform = poseToWorldTransform(convertToPose(laser_offset_));
  return transformPoint(offset_transform, offset_center);
}


optional<gm::Point> ScanManager::nodeBarycenter (const NodeId n) const
{
  if (contains(*optimized_poses_, n) && contains(barycenters_, n)) 
    return optional<gm::Point>(nodeBarycenter(n, keyValue(*optimized_poses_, n)));
  else
    return optional<gm::Point>();
}

gm::Point ScanManager::nodeBarycenter (const NodeId n, const gm::Pose& pose) const
{
  return pointInReferenceFrame(keyValue(barycenters_, n), pose);
}


NodeSet ScanManager::nearbyNodes (const NodeId n, const Point& center, const PoseGraph& g,
    const double max_distance) const
{
  NodeSet nodes;
  
  std::queue<NodeId> q;
  q.push(n);
  ROS_DEBUG_STREAM_NAMED ("scan_manager_nearby_nodes", "Looking for nodes within " << max_distance
                          << " of " << n << " at " << toString(center));

  while (!q.empty()) {
    const NodeId current = q.front();
    optional<gm::Point> current_center = nodeBarycenter(current);
    ROS_ASSERT(current_center);
    q.pop();

    if (!contains(nodes, current) && 
        euclideanDistance(*current_center, center) < max_distance) {
      nodes.insert(current);
      ROS_DEBUG_STREAM_NAMED ("scan_manager_nearby_nodes", "Adding nearby node " <<
                              current << " at " << toString(*current_center));
      BOOST_FOREACH (const NodeId n2, g.neighbors(current))
        q.push(n2);
    }
  }

  return nodes;
}

NodeConstraint ScanManager::scanChainConstraint (const gm::Pose2D& pose, const Chain& chain) const
{
  const NodeId closest = closestNode(pose, chain);
  const Pose2D closest_pose = projectToPose2D(keyValue(*optimized_poses_, closest));
  return NodeConstraint(closest, makeConstraint(relativeTransform(pose, closest_pose)));
}


bool closerTo (const gm::Pose2D& pose, const NodePoseMap& optimized_poses,
               const NodeId n1, const NodeId n2)
{
  const double d1=euclideanDistance(pose, projectToPose2D(keyValue(optimized_poses, n1)));
  const double d2=euclideanDistance(pose, projectToPose2D(keyValue(optimized_poses, n2)));
  return d1<d2;
}


NodeId ScanManager::closestNode (const gm::Pose2D& pose, const Chain& chain) const
{
  Chain::const_iterator pos = min_element (chain.begin(), chain.end(), 
                                           bind(closerTo, pose, *optimized_poses_, _1, _2));
  ROS_ASSERT (pos!=chain.end());
  return *pos;
}

NodeId ScanManager::closestNode (const gm::Pose2D& pose, const NodeSet& nodes) const
{
  NodeSet::const_iterator pos = min_element (nodes.begin(), nodes.end(), 
                                             bind(closerTo, pose, *optimized_poses_, _1, _2));
  ROS_ASSERT (pos!=nodes.end());
  return *pos;
}



ksm::ScanMatchResult ScanManager::scanMatchNodes (KartoMatcherPtr matcher,
                                                  const gm::Pose& init_pose_estimate, 
                                                  const LaserPtr scan, const NodeSet& nodes) const
{
  
  return matcher->scanMatch(*scan, projectToPose2D(init_pose_estimate), makeRefScans(nodes));
}

void ScanManager::setOptimizedPoses (const NodePoseMap& optimized_poses) const 
{
  optimized_poses_ = optimized_poses;
}

void ScanManager::unsetOptimizedPoses () const
{
  optimized_poses_ = optional<NodePoseMap>();
}

WithOptimizedPoses::WithOptimizedPoses (const ScanManager* matcher, const NodePoseMap& poses) :
  matcher(matcher)
{
  matcher->setOptimizedPoses(poses);
}

WithOptimizedPoses::~WithOptimizedPoses ()
{
  matcher->unsetOptimizedPoses();
}

  


}
