/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */


/**
 * \file
 *
 * Implements ScanMatchLocalization
 *
 * \author Bhaskara Marthi
 */

#include <pose_graph/scan_match_localization.h>
#include <pose_graph/utils.h>
#include <pose_graph/transforms.h>

namespace pose_graph
{

namespace gm=geometry_msgs;
namespace ksm=karto_scan_matcher;

ScanMatchLocalization::ScanMatchLocalization (const gm::Pose2D& laser_offset, double nbd_size) :
  ScanManager(laser_offset), max_link_length_(nbd_size)
{}

LocalizationState ScanMatchLocalization::nextState (const LocalizationState& state,
                                                    const gm::Pose& fixed_frame_pose, const LaserPtr scan, 
                                                    const PoseGraph& graph, 
                                                    const NodePoseMap& optimized_poses) const
{
  ROS_DEBUG_STREAM_NAMED ("localization", "Currently at pose " << projectToPose2D(state.node_frame_pose) << 
                          " w.r.t node " << state.ref_node);

  // Initialize matcher if necessary
  if (!matcher_)
    matcher_.reset(new ksm::KartoScanMatcher(*scan, laser_offset_, 0.3, 0.01));
  
  // Set optimized poses during this function
  WithOptimizedPoses opt(this, optimized_poses);
  
  // Identify transform between fixed frame and implicit frame of the optimized poses
  const gm::Pose opt_ref_pose = keyValue(optimized_poses, state.ref_node);
  const Eigen::Transform3d node_to_opt(poseToWorldTransform(opt_ref_pose));
  const Eigen::Transform3d fixed_frame_to_node(state.node_to_fixed_frame.inverse());
  const Eigen::Transform3d fixed_frame_to_opt = node_to_opt*fixed_frame_to_node;
  
  // Initial estimate of current pose in optimized frame
  const gm::Pose init_pose = applyTransform(fixed_frame_to_opt, fixed_frame_pose);
  ROS_DEBUG_STREAM_NAMED ("localization", "Fixed frame pose is " << projectToPose2D(fixed_frame_pose) 
                          << " and optimized frame pose is " << projectToPose2D(init_pose));
  ROS_DEBUG_STREAM_NAMED ("localization", "Fixed frame to opt transform is " << toString2D(fixed_frame_to_opt));

  // Barycenter
  const gm::Point barycenter_base = computeBarycenterInBaseFrame(scan);
  const gm::Point barycenter_fixed = pointInReferenceFrame(barycenter_base, fixed_frame_pose);
  const gm::Point barycenter_opt = transformPoint(fixed_frame_to_opt, barycenter_fixed);
  ROS_DEBUG_STREAM_NAMED ("localization", "Barycenter in fixed frame is " << barycenter_fixed);

  // Nodes that we'll scan match against
  const NodeSet nbd = nearbyNodes(state.ref_node, barycenter_opt, graph, max_link_length_);

  if (nbd.size() > 0) {

    // Do the scan matching
    const ksm::ScanMatchResult res = scanMatchNodes(matcher_, init_pose, scan, nbd);
    const gm::Pose corrected_pose = convertToPose(res.pose);
    ROS_DEBUG_STREAM_NAMED ("localization", "Scan matching returned pose " << res.pose << 
                            " with response " << res.response);
  
    // Figure out the new closest node
    Eigen::Transform3d corrected_fixed_frame_to_opt = transformBetween(fixed_frame_pose, corrected_pose);
    const NodeId closest = closestNode(projectToPose2D(corrected_pose), nbd);
    const gm::Pose closest_pose = keyValue(optimized_poses, closest);
    ROS_DEBUG_STREAM_NAMED ("localization", "Corrected fixed frame to opt transform is " 
                            << toString2D(corrected_fixed_frame_to_opt));
    ROS_DEBUG_STREAM_NAMED ("localization", "New closest node is " << closest << " at " << closest_pose);

    // Return updated state
    const Eigen::Transform3d opt_to_closest(poseToWorldTransform(closest_pose).inverse());
    const Eigen::Transform3d corrected_fixed_frame_to_node=opt_to_closest*corrected_fixed_frame_to_opt;
    const Pose new_local_pose = applyTransform(corrected_fixed_frame_to_node, fixed_frame_pose);
    const Eigen::Transform3d node_to_fixed_frame(corrected_fixed_frame_to_node.inverse());
    return LocalizationState(closest, node_to_fixed_frame, new_local_pose);

  }
  else {
    ROS_WARN ("In localization update, current position had an empty neighborhood, so not updating");
    return state;
  }
}


} // namespace pose_graph
