/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <pose_graph/sequential_scan_matcher.h>
#include <pose_graph/geometry.h>
#include <pose_graph/utils.h>
#include <pose_graph/transforms.h>
#include <boost/foreach.hpp>
#include <ros/ros.h>

namespace pose_graph
{

namespace gm=geometry_msgs;
namespace ksm=karto_scan_matcher;
using std::vector;

SequentialScanMatcher::SequentialScanMatcher (const gm::Pose2D& laser_offset, const double max_link_length,
                                              const unsigned min_chain_length, const double min_response) :
  ScanManager(laser_offset), max_link_length_(max_link_length),
  min_chain_length_(min_chain_length), min_response_(min_response)
{
  ROS_DEBUG_STREAM_NAMED ("init", "Initializing sequential scan matcher with max link length " <<
                          max_link_length_ << ", min chain length " << min_chain_length_ <<
                          ", min response " << min_response_);
}

NodeConstraintVector SequentialScanMatcher::getConstraints (const PoseGraph& graph, const NodeId ref_node,
                                                            const Pose& initial_pose_estimate,
                                                            const LaserPtr scan) const
{
  vector<Chain> chains = candidateChains(graph, ref_node, initial_pose_estimate, scan);
  NodeConstraintVector constraints;

  // Initialize the matcher based on the parameters of \a scan if necessary
  if (!matcher_)
    matcher_.reset(new ksm::KartoScanMatcher(*scan, laser_offset_, 0.3, 0.01));

  for (unsigned i=0; i<chains.size(); i++) {
    NodeSet nodes(chains[i].begin(), chains[i].end());
    ksm::ScanMatchResult result = scanMatchNodes(matcher_, initial_pose_estimate, scan, nodes);
    if (result.response >= min_response_ || i==0) {
      constraints.push_back(scanChainConstraint(result.pose, chains[i]));
    }
  }
  return constraints;
}


void addChain (const Chain& chain, vector<Chain>* chains, NodeSet* processed)
{
  ROS_DEBUG_NAMED ("scan_odom_chains", "Adding chain starting at %u with %zu nodes", chain[0], chain.size());
  chains->push_back(chain);
  BOOST_FOREACH (const NodeId n, chain) {
    ROS_ASSERT(!contains(*processed, n));
    processed->insert(n);
  }
}


ChainVec SequentialScanMatcher::candidateChains (const PoseGraph& g, const NodeId ref_node, 
                                                 const Pose& initial_pose_estimate, 
                                                 LaserPtr scan) const
{
  vector<Chain> chains;
  NodeSet processed;
  const gm::Point barycenter = pointInReferenceFrame(computeBarycenterInBaseFrame(scan), initial_pose_estimate);
  
  ROS_DEBUG_STREAM_NAMED ("scan_odom_chains", "Getting chains near " << barycenter 
                          << " starting at " << ref_node);

  NodeSet nodes = nearbyNodes(ref_node, barycenter, g, max_link_length_);

  BOOST_FOREACH (const NodeId seed, nodes) {
    if (!contains(processed, seed)) {
      Chain candidate = getChain(seed, nodes, processed);
      if (candidate.size() >= min_chain_length_) 
        addChain(candidate, &chains, &processed);
    }
  }
  return chains;
}



} // namespace pose_graph
