#include <pose_graph/spa_2d_conversion.h>
#include <tf/transform_datatypes.h>
#include <boost/foreach.hpp>
#include <set>

/************************************************************
 * NOTE: FILEWIDE DEFINE
 ************************************************************/
#define foreach BOOST_FOREACH




namespace pose_graph
{

using std::set;
using boost::tie;
using sba::Node2d;
using sba::SysSPA2d;
using sba::Con2dP2;
using geometry_msgs::Pose;

// Make a spa node given an initial pose estimate
Node2d makeNode (const double x, const double y, const double theta)
{
  Node2d n;
  n.trans(0) = x;
  n.trans(1) = y;
  n.trans(2) = 1.0;
  n.arot = theta;
  n.setTransform();
  n.setDr();
  return n;
}


// Get yaw from an eigen quaternion
double getYaw (const Quaterniond& q)
{
  const tf::Quaternion qt(q.x(), q.y(), q.z(), q.w());
  return tf::getYaw(qt);
}

Con2dP2 makeSpaConstraint (const PoseConstraint& constraint)
{
  Con2dP2 spa;
  spa.tmean[0] = constraint.translation[0];
  spa.tmean[1] = constraint.translation[1];
  spa.amean = getYaw(constraint.rotation);

  // Todo: set precision correctly based on constraint.precision
  spa.prec.setIdentity();

  return spa;
}


// Turn a pose graph into a pointer to a Spa2d
SpaConversionResult poseGraphToSpa (const PoseGraph& pose_graph)
{
  SpaPtr spa(new SysSPA2d());
  map<NodeId, unsigned> node_indices;
  unsigned next_index=0;

  // 1. Add the nodes
  foreach (const NodeId n, pose_graph.allNodes()) {
    spa->nodes.push_back(makeNode(0.0, 0.0, 0.0));
    node_indices[n] = next_index++;
  }



  // 2. Add the edge constraints
  foreach (const EdgeId e, pose_graph.allEdges()) {
    
    NodeId from, to;
    tie(from, to) = pose_graph.incidentNodes(e);
    const pose_graph::PoseConstraint constraint = pose_graph.getConstraint(e);
    Con2dP2 spa_constraint = makeSpaConstraint(constraint);

    spa_constraint.ndr = node_indices[from];
    spa_constraint.nd1 = node_indices[to];
    spa->p2cons.push_back(spa_constraint);
    ROS_DEBUG_STREAM_NAMED ("pose_graph_spa", "Adding constraint with tmean " << spa_constraint.tmean << 
                            " and amean " << spa_constraint.amean << " from " << spa_constraint.ndr << 
                            " to " << spa_constraint.nd1);
  }

  
  return SpaConversionResult(spa, node_indices);
}


// Return the Pose corresponding to a spa node
Pose getNodePose (const sba::Node2d& n)
{
  Pose pose;

  pose.position.x = n.trans(0);
  pose.position.y = n.trans(1);
  pose.position.z = 0.0;
  
  pose.orientation = tf::createQuaternionMsgFromYaw(n.arot);

  return pose;
}





} // namespace
